# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import datastock as ds


# tofu
# from tofu import __version__ as __version__
from . import _class10_refs as _refs
from . import _class8_plot


# ################################################################
# ################################################################
#                           inversions
# ################################################################


def _plot_inversion_check(
    coll=None,
    key=None,
    plot_details=None,
    indbf=None,
    indchan=None,
    dvminmax=None,
    cmap=None,
    dcolorbar=None,
    dleg=None,
    alpha=None,
    # los sampling
    los_res=None,
    # interactivity
    color_dict=None,
    nlos=None,
    dax=None,
    connect=None,
):

    # key
    if 'inversions' not in coll.dobj.keys():
        msg = 'No inversions available!'
        raise Exception(msg)

    lk = list(coll.dobj['inversions'].keys())
    keyinv = ds._generic_check._check_var(
        key, 'key',
        default=None,
        types=str,
        allowed=lk,
    )

    wm = coll._which_mesh
    wbs = coll._which_bsplines
    keymat = coll.dobj['inversions'][keyinv]['matrix']
    key_data = coll.dobj['inversions'][keyinv]['data_in']
    key_retro = coll.dobj['inversions'][keyinv]['retrofit']
    keybs = coll.dobj['geom matrix'][keymat]['bsplines']
    key_diag = coll.dobj['geom matrix'][keymat]['diagnostic']
    is2d = coll.dobj['diagnostic'][key_diag]['is2d']
    key_cam = coll.dobj['geom matrix'][keymat]['camera']
    key_retro = coll.dobj['synth sig'][key_retro]['data']
    keym = coll.dobj[wbs][keybs]['mesh']
    mtype = coll.dobj[wm][keym]['type']
    nd = coll.dobj[wm][keym]['nd']
    # refbs = coll.dobj['bsplines'][keybs]['ref']

    crop = coll.dobj['geom matrix'][keymat]['crop']
    if crop is True:
        cropbs = coll.dobj['bsplines'][keybs]['crop']
        cropbs = coll.ddata[cropbs]['data']
    else:
        cropbs = None

    # cmap
    if cmap is None:
        cmap = 'viridis'

    # dcolorbar
    defdcolorbar = {
        # 'location': 'right',
        'fraction': 0.15,
        'orientation': 'vertical',
    }
    dcolorbar = ds._generic_check._check_var(
        dcolorbar, 'dcolorbar',
        default=defdcolorbar,
        types=dict,
    )

    # plot_details
    plot_details = ds._generic_check._check_var(
        plot_details, 'plot_details',
        types=bool,
        default=True,
    )

    # los_res
    los_res = ds._generic_check._check_var(
        los_res, 'los_res',
        types=float,
        default=0.05,
        sign='> 0.',
    )

    # color_dict
    color_dict = _class8_plot._check_color_dict(color_dict)


    # alpha
    alpha = ds._generic_check._check_var(
        alpha, 'alpha',
        types=float,
        default=0.2,
        sign='> 0.',
    )

    # nlos
    nlos = ds._generic_check._check_var(
        nlos, 'nlos',
        types=int,
        default=5,
    )

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = ds._generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    # connect
    connect = ds._generic_check._check_var(
        connect, 'connect',
        default=True,
        types=bool,
    )

    return (
        keyinv, keymat,
        key_diag, key_cam, keybs, key_data, key_retro,
        is2d, mtype, nd,
        cropbs, cmap, dcolorbar,
        nlos, los_res, color_dict, alpha,
        dleg, plot_details, connect,
    )


def _plot_inversion_prepare(
    coll=None,
    coll2=None,
    is2d=None,
    mtype=None,
    keyinv=None,
    key_matrix=None,
    key_diag=None,
    key_cam=None,
    key_data=None,
    key_retro=None,
    los_res=None,
    dref_vector=None,
    dx0=None,
    dx1=None,
):

    # ------------
    # check

    if dref_vector is None:
        dref_vector = {}

    # -----------------
    # add nearest-neighbourg interpolated data

    # just for preparation
    ddata = {
        k0: {'data': coll.ddata[key_data[ii]]} for ii, k0 in enumerate(key_cam)
    }

    # dcamref
    dcamref, drefx, drefy = _class8_plot._prepare_dcamref(
        coll=coll,
        key_cam=key_cam,
        is2d=is2d,
    )

    # los
    dlos_n, dref_los = _class8_plot._prepare_los(
        coll=coll,
        coll2=coll2,
        dcamref=dcamref,
        key_diag=key_diag,
        key_cam=key_cam,
        los_res=los_res,
    )

    # vos
    dvos_n, dref_vos = _class8_plot._prepare_vos(
        coll=coll,
        coll2=coll2,
        dcamref=dcamref,
        key_diag=key_diag,
        key_cam=key_cam,
        los_res=los_res,
    )

    # ddatax, ddatay
    (
        _, dkeyx, dkeyy, ddatax, ddatay, dextent,
    ) = _class8_plot._prepare_datarefxy(
        coll=coll,
        coll2=coll2,
        dcamref=dcamref,
        drefx=drefx,
        drefy=drefy,
        ddata=ddata,
        is2d=is2d,
        dx0=dx0,
        dx1=dx1,
    )

    # -----------
    # get reft

    hastime, reft, keyt, t, dind = _refs._get_ref_vector_common(
        coll=coll,
        key_matrix=key_matrix,
        key_profile2d=keyinv,
        dref_vector=dref_vector,
    )

    # -----------------
    # add nearest-neighbourg interpolated data

    reft, keyt, time = coll.get_ref_vector(
        key0=keyinv,
        ref=reft,
        **dref_vector,
    )[2:5]
    lkmat = coll.dobj['geom matrix'][key_matrix]['data']

    dind = None
    if coll.get_time(key=lkmat[0])[0]:
        keyt_data = coll.get_time(key=key_data[0])[3]
        if keyt_data != keyt:
            dind = coll.get_time(
                key=key_data[0],
                t=keyt,
            )[-1]

    if reft not in coll2.dref.keys():
        coll2.add_ref(key=reft, size=coll.dref[reft]['size'])

    # --------------------
    # add data, retro, err

    datamin, datamax = np.inf, -np.inf
    errmin, errmax = np.inf, -np.inf
    for ii, k0 in enumerate(key_cam):

        # ref
        for rr in coll.dobj['camera'][k0]['dgeom']['ref']:
            if rr not in coll2.dref.keys():
                coll2.add_ref(key=rr, size=coll.dref[rr]['size'])

        refi = coll.ddata[key_data[ii]]['ref']
        datai = coll.ddata[key_data[ii]]['data']
        if dind is not None:
            nd = len(refi)
            refti = coll.get_time(key=key_data[ii])[2]
            axis = refi.index(refti)
            refi = tuple([reft if jj == axis else refi[jj] for jj in range(nd)])
            sli = tuple([
                dind['ind'] if jj == axis else slice(None) for jj in range(nd)
            ])
            datai = datai[sli]

        # transpose for 2d
        retroi = coll.ddata[key_retro[ii]]['data']
        if is2d:
            ndim = datai.ndim
            datai = np.swapaxes(datai, ndim-1, ndim-2)
            retroi = np.swapaxes(retroi, ndim-1, ndim-2)
            refi = list(refi)
            refi[-2], refi[-1] = refi[-1], refi[-2]
            refi = tuple(refi)

        # data min max
        datamin = min(datamin, np.nanmin(datai))
        datamax = max(datamax, np.nanmax(datai))

        # data
        coll2.add_data(
            key=key_data[ii],
            data=datai,
            ref=refi,
            units=coll.ddata[key_data[ii]]['units'],
        )

        # retro
        coll2.add_data(
            key=key_retro[ii],
            data=retroi,
            ref=refi,
        )

        # err min max
        erri = retroi - datai
        errmin = min(errmin, np.nanmin(erri))
        errmax = max(errmax, np.nanmax(erri))

        # data
        coll2.add_data(
            key=f"{key_data[ii]}_err",
            data=erri,
            ref=refi,
            units=coll.ddata[key_data[ii]]['units'],
        )

        # cross-section los

    # errmax
    errmax = max(np.abs(errmin), np.abs(errmax))

    # ----------------
    # inversion parameters

    if reft is not None:
        chi2n = coll.ddata[f'{keyinv}_chi2n']['data']
        mu = coll.ddata[f'{keyinv}_mu']['data']
        reg = coll.ddata[f'{keyinv}_reg']['data']
        niter = coll.ddata[f'{keyinv}_niter']['data']
    else:
        chi2n = None    # coll.dobj['inversions'][keyinv]['chi2n']
        mu = None       # coll.dobj['inversions'][keyinv]['mu']
        reg = None      # coll.dobj['inversions'][keyinv]['reg']
        niter = None    # coll.dobj['inversions'][keyinv]['niter']

    return (
        dlos_n, dref_los,
        dvos_n, dref_vos,
        drefx, drefy, dkeyx, dkeyy, ddatax, ddatay, dextent,
        time, keyt, reft,
        chi2n, mu, reg, niter,
        datamin, datamax, errmax,
    )


def plot_inversion(
    coll=None,
    key=None,
    plot_details=None,
    res=None,
    dvminmax=None,
    cmap=None,
    alpha=None,
    # config
    plot_config=None,
    # figure
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
    # los sampling
    los_res=None,
    # ref vector specifier
    dref_vector=None,
    ref_vector_strategy=None,
    # offset
    dx0=None,
    dx1=None,
    # interactivity
    color_dict=None,
    nlos=None,
    dinc=None,
    connect=None,
):

    # ------------
    # check inputs

    (
        keyinv, keymat,
        key_diag, key_cam, keybs, key_data, key_retro,
        is2d, mtype, nd,
        cropbs, cmap, dcolorbar,
        nlos, los_res, color_dict, alpha,
        dleg, plot_details, connect,
    ) = _plot_inversion_check(
        coll=coll,
        key=key,
        plot_details=plot_details,
        cmap=cmap,
        dcolorbar=dcolorbar,
        dleg=dleg,
        alpha=alpha,
        # los sampling
        los_res=los_res,
        # interactivity
        color_dict=color_dict,
        nlos=nlos,
        connect=connect,
    )

    # --------------
    # plot - prepare

    if dax is None:

        dax = _plot_inversion_create_axes(
            fs=fs,
            dmargin=dmargin,
            nd=nd,
            key_cam=key_cam,
        )

    dax = ds._generic_check._check_dax(dax=dax, main='matrix')

    # --------------
    # plot profile2d

    coll2, dgroup = coll.plot_as_profile2d(
        key=keyinv,
        dres=res,
        plot_details=plot_details,
        # ref vectors
        dref_vectorZ=dref_vector,
        dref_vectorU=None,          # U not handled yet
        ref_vector_strategy=ref_vector_strategy,
        # figure
        dvminmax=dvminmax,
        cmap=cmap,
        dax=dax,
        dmargin=dmargin,
        fs=fs,
        dcolorbar=dcolorbar,
        dleg=dleg,
        # interactivity
        dinc=dinc,
        connect=False,
    )
    dax = coll2.dax

    # ------------
    # prepare data

    (
        dlos_n, dref_los,
        dvos_n, dref_vos,
        drefx, drefy, dkeyx, dkeyy, ddatax, ddatay, dextent,
        time, keyt, reft,
        chi2n, mu, reg, niter,
        datamin, datamax, errmax,
    ) = _plot_inversion_prepare(
        coll=coll,
        coll2=coll2,
        is2d=is2d,
        keyinv=keyinv,
        key_matrix=keymat,
        key_cam=key_cam,
        key_diag=key_diag,
        key_data=key_data,
        key_retro=key_retro,
        los_res=los_res,
        dref_vector=dref_vector,
        dx0=dx0,
        dx1=dx1,
    )

    # ----------------
    # define and set dgroup

    dgroup.update({
        f'{k0}_i0': {
            'ref': [drefx[k0]],
            'data': ['index'],
            'nmax': nlos,
        }
        for k0 in key_cam
    })

    if is2d:
        dgroup.update({
            f'{k0}_i1': {
                'ref': [drefy[k0]],
                'data': ['index'],
                'nmax': nlos,
            }
            for k0 in key_cam
        })

    if reft is not None and reft not in dgroup['Z']['ref']:
        dgroup['Z']['ref'].append(reft)

    # ---------
    # plot data

    for ii, k0 in enumerate(key_cam):

        # data vs retro
        kax = k0
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            if is2d:

                if reft is None:
                    ax.imshow(
                        coll2.ddata[key_data[ii]]['data'],
                        extent=dextent[k0],
                        cmap=cmap,
                        # vmin=vmin,
                        # vmax=vmax,
                        origin='lower',
                        interpolation='nearest',
                        aspect='equal',
                    )

                else:

                    im = ax.imshow(
                        coll2.ddata[key_data[ii]]['data'][0, ...],
                        extent=dextent[k0],
                        cmap=cmap,
                        # vmin=vmin,
                        # vmax=vmax,
                        origin='lower',
                        interpolation='nearest',
                        aspect='equal',
                    )

                    ki = key_data[ii]
                    coll2.add_mobile(
                        key=ki,
                        handle=im,
                        refs=(reft,),
                        data=[key_data[ii]],
                        dtype=['data'],
                        axes=kax,
                        ind=0,
                    )

            else:

                nch = coll.dobj['camera'][k0]['dgeom']['pix_nb']
                chani = np.arange(0, nch)
                if reft is None:
                    ax.plot(
                        chani,
                        coll2.ddata[key_data[ii]]['data'].ravel(),
                        c=(0.8, 0.8, 0.8),
                        marker='o',
                        ls='-',
                    )

                    ax.plot(
                        chani,
                        coll2.ddata[key_retro[ii]]['data'].ravel(),
                        c='k',
                        marker='None',
                        ls='-',
                        lw=2,
                    )

                else:
                    nani = np.nan * chani
                    l0, = ax.plot(
                        chani,
                        nani,
                        c=(0.8, 0.8, 0.8),
                        marker='o',
                        ls='-',
                    )

                    l1, = ax.plot(
                        chani,
                        nani,
                        c='k',
                        marker='None',
                        ls='-',
                        lw=2,
                    )

                    kl0 = key_data[ii]
                    coll2.add_mobile(
                        key=kl0,
                        handle=l0,
                        refs=(reft,),
                        data=[key_data[ii]],
                        dtype=['ydata'],
                        axes=kax,
                        ind=0,
                    )

                    kl1 = key_retro[ii]
                    coll2.add_mobile(
                        key=kl1,
                        handle=l1,
                        refs=(reft,),
                        data=[key_retro[ii]],
                        dtype=['ydata'],
                        axes=kax,
                        ind=0,
                    )
                    ax.set_ylim(min(0, datamin), datamax)
                    ax.set_xlim(0, nch + 1)

            # add vlines / markers
            _class8_plot._add_camera_vlines_marker(
                coll2=coll2,
                dax=dax,
                ax=ax,
                kax=kax,
                is2d=is2d,
                k0=k0,
                nlos=nlos,
                ddatax=ddatax,
                ddatay=ddatay,
                drefx=drefx,
                drefy=drefy,
                dkeyx=dkeyx,
                dkeyy=dkeyy,
                color_dict=color_dict,
            )

        # add los
        kax = 'matrix'
        if dlos_n[k0] is not None:
            ax = dax[kax]['handle']

            nan_los = np.full((dlos_n[k0],), np.nan)
            if dvos_n is None:
                nan_vos = None
            else:
                nan_vos = np.full((dvos_n[k0],), np.nan)

            _class8_plot._add_camera_los_cross(
                coll2=coll2,
                k0=k0,
                ax=ax,
                kax=kax,
                nlos=nlos,
                dref_los=dref_los,
                dref_vos=dref_vos,
                color_dict=color_dict,
                nan_los=nan_los,
                nan_vos=nan_vos,
                alpha=alpha,
            )

        # err
        kax = f"{k0}_err"
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            kerr = f"{key_data[ii]}_err"
            if is2d:

                if reft is None:
                    ax.imshow(
                        coll2.ddata[kerr]['data'],
                        extent=dextent[k0],
                        cmap=cmap,
                        # vmin=vmin,
                        # vmax=vmax,
                        origin='lower',
                        interpolation='nearest',
                        aspect='equal',
                    )

                else:

                    im = ax.imshow(
                        coll2.ddata[kerr]['data'][0, ...],
                        extent=dextent[k0],
                        cmap=cmap,
                        # vmin=vmin,
                        # vmax=vmax,
                        origin='lower',
                        interpolation='nearest',
                        aspect='equal',
                    )

                    ki = kerr
                    coll2.add_mobile(
                        key=ki,
                        handle=im,
                        refs=(reft,),
                        data=[kerr],
                        dtype=['data'],
                        axes=kax,
                        ind=0,
                    )

            else:

                nch = coll.dobj['camera'][k0]['dgeom']['pix_nb']
                chani = np.arange(0, nch)
                if reft is None:
                    ax.plot(
                        chani,
                        coll2.ddata[kerr]['data'].ravel(),
                        c=(0.8, 0.8, 0.8),
                        marker='o',
                        ls='-',
                    )

                else:
                    nani = np.nan * chani
                    l0, = ax.plot(
                        chani,
                        nani,
                        c=(0.8, 0.8, 0.8),
                        marker='o',
                        ls='-',
                    )

                    kl0 = kerr
                    coll2.add_mobile(
                        key=kl0,
                        handle=l0,
                        refs=(reft,),
                        data=[kerr],
                        dtype=['ydata'],
                        axes=kax,
                        ind=0,
                    )

                    ax.set_ylim(-errmax, errmax)
                    ax.set_xlim(0, nch + 1)

            # add vlines / markers
            _class8_plot._add_camera_vlines_marker(
                coll2=coll2,
                dax=dax,
                ax=ax,
                kax=kax,
                is2d=is2d,
                k0=k0,
                nlos=nlos,
                ddatax=ddatax,
                ddatay=ddatay,
                drefx=drefx,
                drefy=drefy,
                dkeyx=dkeyx,
                dkeyy=dkeyy,
                color_dict=color_dict,
                suffix='_err'
            )

    # # -------------------------
    # # plot inversion parameters

    if reft is not None:
        kax = 'inv-param'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            ax.plot(
                time,
                chi2n / np.nanmax(chi2n),
                c='k',
                ls='-',
                lw=1.,
                marker='.',
                label='chi2n norm',
            )

            ax.plot(
                time,
                reg / np.nanmax(reg),
                c='b',
                ls='-',
                lw=1.,
                marker='.',
                label='mu*reg norm',
            )

            # add mobile
            l0 = ax.axvline(time[0], c='k', ls='-', lw=1.)

            # add mobile
            kl0 = 't-par'
            coll2.add_mobile(
                key=kl0,
                handle=l0,
                refs=(reft,),
                data=[keyt],
                dtype=['xdata'],
                axes=kax,
                ind=0,
            )
            ax.set_ylim(bottom=0)

            ax.legend()

        kax = 'niter'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            ax.plot(
                time,
                niter,
                c='k',
                ls='-',
                lw=1.,
                marker='.',
            )

            # add mobile
            l0 = ax.axvline(time[0], c='k', ls='-', lw=1.)

            # add mobile
            kl0 = 't-niter'
            coll2.add_mobile(
                key=kl0,
                handle=l0,
                refs=(reft,),
                data=[keyt],
                dtype=['xdata'],
                axes=kax,
                ind=0,
            )
            ax.set_ylim(bottom=0)

            ax.set_ylim(bottom=0)

    # -------
    # config

    if plot_config.__class__.__name__ == 'Config':

        kax = 'matrix'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj='cross', dLeg=False)

    # -------
    # connect

    if connect is True:
        coll2.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)
        coll2.disconnect_old()
        coll2.connect()

        coll2.show_commands()
        return coll2
    else:
        return coll2, dgroup


def _plot_inversion_create_axes(
    fs=None,
    dmargin=None,
    nd=None,
    key_cam=None,
):

    if fs is None:
        fs = (16, 10)

    if dmargin is None:
        dmargin = {
            'left': 0.05, 'right': 0.98,
            'bottom': 0.05, 'top': 0.95,
            'hspace': 1.5, 'wspace': 0.5,
        }

    ncam = len(key_cam)
    nblock = 3*ncam
    nrows = 3*nblock

    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(ncols=8, nrows=nrows, **dmargin)

    # ------------------
    # axes for profile2d

    # axes for image
    ax0 = fig.add_subplot(gs[:2*nblock, 2:4], aspect='auto')

    # axes for vertical profile
    ax1 = fig.add_subplot(gs[:2*nblock, 4], sharey=ax0)

    # axes for horizontal profile
    ax2 = fig.add_subplot(gs[2*nblock:, 2:4], sharex=ax0)

    # axes for radius
    if nd == '1d':
        ax7 = fig.add_subplot(gs[:nblock, :2], sharey=ax2)
    else:
        ax7 = None

    # axes for traces
    ax3 = fig.add_subplot(gs[nblock:2*nblock, :2])

    # dax
    dax = {
        # data
        'matrix': {'handle': ax0},
        'vertical': {'handle': ax1},
        'horizontal': {'handle': ax2},
        'tracesZ': {'handle': ax3},
    }
    # axes for text
    # ax4 = fig.add_subplot(gs[:3, 5], frameon=False)
    # ax5 = fig.add_subplot(gs[3:, 5], frameon=False)
    # ax6 = fig.add_subplot(gs[4:, :2], frameon=False)

    # ------------------
    # axes for inversion

    npc = nrows / ncam
    for ii, k0 in enumerate(key_cam):
        # retrofit
        dax[k0] = fig.add_subplot(gs[9*ii:9*ii+6, 5:])

        # error
        dax[f'{k0}_err'] = fig.add_subplot(gs[9*ii+6:9*(ii+1), 5:], sharex=dax[k0])

    # parameters (chi2, ...)
    dax['inv-param'] = fig.add_subplot(gs[2*nblock:2*nblock+ncam, :2], sharex=ax3)

    # nb of iterations
    dax['niter'] = fig.add_subplot(gs[2*nblock+ncam:, :2], sharex=ax3)

    if ax7 is not None:
        dax['radial'] = {'handle': ax7}

    return dax


# Common
import numpy as np
import matplotlib.pyplot as plt

# ToFu-specific



def plot(dust, lax=None, Proj='All'):

    if dust.Ves is not None:
        lax = dust.Ves.plot(Lax=lax, Elt='P', Proj=Proj)
    if dust.LStruct is not None:
        for ss in dust.LStruct:
            lax = ss.plot(Lax=lax, Elt='P', Proj=Proj)

    if dust.traj is not None:
        lax = _plot_traj(dust, lax=lax, Proj=Proj)





def _plot_traj(dust, lax=None, Proj='All'):

    if dust.Type=='Tor':
        ptsCross = np.r_[np.hypot(dust.traj['pts'][0,:],dust.traj['pts'][1,:]),
                         dust.traj['pts'][2,:]]
    else:
        ptsCross = dust.traj['pts'][1:,:]
    ptsHor = dust.traj['pts'][:2,:]

    if lax[0] is not None:
        lax[0].plot(ptsCross[0,:],ptsCross[1,:], c='k', lw=1., ls='-')
    if lax[1] is not None:
        lax[1].plot(ptsHor[0,:],ptsHor[1,:], c='k', lw=1., ls='-')
    return lax



# -*- coding: utf-8 -*-


# Built-in
# import copy


# Common
import numpy as np
import datastock as ds


# tofu
from ._class03_Aperture import Aperture as Previous
from . import _class3_check
from . import _class4_check as _check


__all__ = ['Filter']


# #############################################################################
# #############################################################################
#                           Diagnostic
# #############################################################################


class Filter(Previous):

    _dshow = dict(Previous._dshow)
    _dshow.update({
        'filter': [
            'dgeom.type',
            'dgeom.curve_r',
            'dgeom.area',
            'dmat.name',
            'dgeom.outline',
            'dgeom.poly',
            'dgeom.cent',
            'dmat.energy',
            'dmat.transmission',
            # 'dmisc.color',
        ],
    })

    def add_filter(
        self,
        key=None,
        # geometry
        dgeom=None,
        # material
        dmat=None,
        # dmisc
        color=None,
    ):
        """ Add a filter

        dgeom is a dict holding the geometry:
            - 2d outline + 3d center + unit vectors (nin, e0, e1)


        dmat is a dict holding the material properties


        Unit vectors will be checked and normalized
        If planar, area will be computed
        Outline will be made counter-clockwise

        """

        # check / format input
        dref, ddata, dobj = _class3_check._add_surface3d(
            coll=self,
            key=key,
            which='filter',
            which_short='filt',
            # 2d outline
            **dgeom,
        )

        key = list(dobj['filter'].keys())[0]

        # material
        dref2, ddata2, dmat = _check._dmat(
            key=key,
            dmat=dmat,
        )

        if dmat is not None:
            dref.update(dref2)
            ddata.update(ddata2)
            dobj['filter'][key]['dmat'] = dmat

        # dmisc
        dobj['filter'][key]['dmisc'] = _class3_check._dmisc(
            key=key,
            color=color,
        )

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)





import numpy as np
import datastock as ds


# #################################################################
# #################################################################
#                   Get knots per bsplines 1d
# #################################################################


def _get_knots_per_bs(
    knots,
    deg=None,
    returnas=None,
    return_unique=None,
    poloidal=None,
):

    # ----------
    # check input

    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        types=str,
        default='data',
        allowed=['ind', 'data'],
    )
    return_unique = ds._generic_check._check_var(
        return_unique, 'return_unique',
        types=bool,
        default=False,
    )

    poloidal = ds._generic_check._check_var(
        poloidal, 'poloidal',
        default=False,
        types=bool,
    )

    # --------
    # prepare

    nkpbs = 2 + deg
    size = knots.size

    if poloidal is True:
        if size < nkpbs - 1:
            msg = (
                f"For the desired degree ({deg}), "
                f"a minimum of {nkpbs - 1} poloidal knots is necessary\n"
                f"Provided: {knots}"
            )
            raise Exception(msg)
        nbs = size
        if deg == 0 and size == nkpbs - 1:
            msg = "Using 2 pts for a deg = 0 bsplines leads to bspline!"
            raise Exception(msg)

    else:
        if size < 1 - deg:
            msg = (
                "For the desired degree ({deg}), "
                f"a minimum of {2 - deg} poloidal knots is necessary\n"
                f"Provided: {knots}"
            )
            raise Exception(msg)
        nbs = size - 1 + deg

    # ----------
    # compute

    if return_unique:
        if deg == 0:
            knots_per_bs = np.arange(0, size)
        elif deg == 1:
            knots_per_bs = np.r_[0, np.arange(0, size), size-1]
        elif deg == 2:
            knots_per_bs = np.r_[0, 0, np.arange(0, size), size-1, size-1]
        elif deg == 3:
            knots_per_bs = np.r_[
                0, 0, 0, np.arange(0, size), size-1, size-1, size-1,
            ]

    else:
        knots_per_bs = np.zeros((nkpbs, nbs), dtype=int)

        if deg == 0:
            if poloidal is True:
                knots_per_bs[:, :] = np.array([
                    np.arange(0, size),
                    np.r_[np.arange(1, size), 0],
                ])
            else:
                knots_per_bs[:, :] = np.array([
                    np.arange(0, size-1),
                    np.arange(1, size),
                ])

        elif deg == 1:
            if poloidal is True:
                if size == nkpbs - 1:
                    knots_per_bs[:, :] = np.array([
                        np.r_[0, 1],
                        np.r_[1, 0],
                        np.r_[0, 1],
                    ])
                else:
                    knots_per_bs[:, :] = np.array([
                        np.arange(0, size),
                        np.r_[np.arange(1, size), 0],
                        np.r_[np.arange(2, size), 0, 1],
                    ])
            else:
                knots_per_bs[:, 1:-1] = np.array([
                    np.arange(0, size-2),
                    np.arange(1, size-1),
                    np.arange(2, size),
                ])
                knots_per_bs[:, 0] = [0, 0, 1]
                knots_per_bs[:, -1] = [-2, -1, -1]

        elif deg == 2:
            if poloidal is True:
                if size == nkpbs - 1:
                    knots_per_bs[:, :] = np.array([
                        np.arange(0, size),
                        np.r_[np.arange(1, size), 0],
                        np.r_[np.arange(2, size), 0, 1],
                        np.arange(0, size),
                    ])
                else:
                    knots_per_bs[:, :] = np.array([
                        np.arange(0, size),
                        np.r_[np.arange(1, size), 0],
                        np.r_[np.arange(2, size), 0, 1],
                        np.r_[np.arange(3, size), 0, 1, 2],
                    ])
            else:
                knots_per_bs[:, 2:-2] = np.array([
                    np.arange(0, size-3),
                    np.arange(1, size-2),
                    np.arange(2, size-1),
                    np.arange(3, size),
                ])
                knots_per_bs[:, 0] = [0, 0, 0, 1]
                knots_per_bs[:, 1] = [0, 0, 1, 2]
                knots_per_bs[:, -2] = [-3, -2, -1, -1]
                knots_per_bs[:, -1] = [-2, -1, -1, -1]

        elif deg == 3:
            if poloidal is True:
                if size == nkpbs - 1:
                    knots_per_bs[:, :] = np.array([
                        np.arange(0, size),
                        np.r_[np.arange(1, size), 0],
                        np.r_[np.arange(2, size), 0, 1],
                        np.r_[np.arange(3, size), 0, 1, 2],
                        np.arange(0, size),
                    ])
                else:
                    knots_per_bs[:, :] = np.array([
                        np.arange(0, size),
                        np.r_[np.arange(1, size), 0],
                        np.r_[np.arange(2, size), 0, 1],
                        np.r_[np.arange(3, size), 0, 1, 2],
                        np.r_[np.arange(4, size), 0, 1, 2, 3],
                    ])
            else:
                knots_per_bs[:, 3:-3] = np.array([
                    np.arange(0, size-4),
                    np.arange(1, size-3),
                    np.arange(2, size-2),
                    np.arange(3, size-1),
                    np.arange(4, size),
                ])
                knots_per_bs[:, 0] = [0, 0, 0, 0, 1]
                knots_per_bs[:, 1] = [0, 0, 0, 1, 2]
                knots_per_bs[:, 2] = [0, 0, 1, 2, 3]
                knots_per_bs[:, -3] = [-4, -3, -2, -1, -1]
                knots_per_bs[:, -2] = [-3, -2, -1, -1, -1]
                knots_per_bs[:, -1] = [-2, -1, -1, -1, -1]

    # ----------
    # return

    if returnas == 'data':
        knots_per_bs = knots[knots_per_bs]

    if return_unique:
        return knots_per_bs, nbs
    else:
        return knots_per_bs


# #################################################################
# #################################################################
#                   Get cents per bsplines 1d
# #################################################################


def _get_cents_per_bs(
    cents,
    deg=None,
    returnas=None,
    poloidal=None,
):

    # ------------
    # check inputs

    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        types=str,
        default='data',
        allowed=['ind', 'data'],
    )

    poloidal = ds._generic_check._check_var(
        poloidal, 'poloidal',
        default=False,
        types=bool,
    )

    # -------
    # prepare

    nkpbs = 1 + deg
    size = cents.size
    nbs = size + deg
    cents_per_bs = np.zeros((nkpbs, nbs), dtype=int)

    # -------
    # compute

    if deg == 0:
        cents_per_bs[0, :] = np.arange(0, size)

    elif deg == 1:
        cents_per_bs[:, 1:-1] = np.array([
            np.arange(0, size-1),
            np.arange(1, size),
        ])
        cents_per_bs[:, 0] = [0, 0]
        cents_per_bs[:, -1] = [-1, -1]

    elif deg == 2:
        cents_per_bs[:, 2:-2] = np.array([
            np.arange(0, size-2),
            np.arange(1, size-1),
            np.arange(2, size),
        ])
        cents_per_bs[:, 0] = [0, 0, 0]
        cents_per_bs[:, 1] = [0, 0, 1]
        cents_per_bs[:, -2] = [-2, -1, -1]
        cents_per_bs[:, -1] = [-1, -1, -1]

    elif deg == 3:
        cents_per_bs[:, 3:-3] = np.array([
            np.arange(0, size-3),
            np.arange(1, size-2),
            np.arange(2, size-1),
            np.arange(3, size),
        ])
        cents_per_bs[:, 0] = [0, 0, 0, 0]
        cents_per_bs[:, 1] = [0, 0, 0, 1]
        cents_per_bs[:, 2] = [0, 0, 1, 2]
        cents_per_bs[:, -3] = [-3, -2, -1, -1]
        cents_per_bs[:, -2] = [-2, -1, -1, -1]
        cents_per_bs[:, -1] = [-1, -1, -1, -1]

    # ------
    # return

    if returnas == 'data':
        cents_per_bs = cents[cents_per_bs]

    return cents_per_bs


# #################################################################
# #################################################################
#                   Get apex positions per bsplines 1d
# #################################################################


def _get_apex_per_bs(
    knots=None,
    knots_per_bs=None,
    deg=None,
    poloidal=None,
):

    # -------
    # prepare

    poloidal = ds._generic_check._check_var(
        poloidal, 'poloidal',
        default=False,
        types=bool,
    )

    nkpbs, nbs = knots_per_bs.shape

    # -------------
    # compute basis

    if nkpbs % 2 == 0:
        ii = int(nkpbs/2)
        apex = np.mean(knots_per_bs[ii-1:ii+1, :], axis=0)

    else:
        ii = int((nkpbs-1) / 2)
        apex = knots_per_bs[ii, :]

    # ------
    # adjust

    if poloidal is True:
        # make sure in [-pi; pi[
        apex = np.arctan2(np.sin(apex), np.cos(apex))
    else:
        # manage edges
        if deg == 1:
            apex[:deg] = knots[0]
            apex[-deg:] = knots[-1]
        elif deg == 2:
            apex[:deg] = [knots[0], 0.5*(knots[0] + knots[1])]
            apex[-deg:] = [0.5*(knots[-2] + knots[-1]), knots[-1]]
        elif deg == 3:
            apex[:deg] = [knots[0], 0.5*(knots[0]+knots[1]), knots[1]]
            apex[-deg:] = [knots[-2], 0.5*(knots[-2]+knots[-1]), knots[-1]]

    return apex


# #################################################################
# #################################################################
#                   Get apex positions per bsplines 1d
# #################################################################


def _coefs_axis(shapebs=None, coefs=None, axis=None, shapex=None):
    
    # -------------------------------------------------
    # initial safety check on coefs vs shapebs vs axis
    
    if np.isscalar(coefs):
        coefs = np.full(shapebs, coefs)
    
    c0 = (
        isinstance(coefs, np.ndarray)
        and coefs.ndim >= len(shapebs)
        and len(axis) == len(shapebs)
        and all([aa == axis[0] + ii for ii, aa in enumerate(axis)])
        and all([coefs.shape[aa] == shapebs[ii] for ii, aa in enumerate(axis)])
    )
    if not c0:
        msg = (
            f"Arg coefs must have a shape including {shapebs}\n"
            f"\t- shapebs: {shapebs}\n"
            f"\t- coefs.shape: {coefs.shape}\n"
            f"\t- axis: {axis}\n"
        )
        raise Exception(msg)
    
    # ----------------
    # shape for output 
    
    shape_pts, axis_x, ind_coefs, ind_x = [], [], [], []
    jj = 0
    for ii in range(coefs.ndim):
        if ii == axis[0]:
            for jj in range(len(shapex)):
                shape_pts.append(shapex[jj])
                axis_x.append(ii + jj)
                ind_x.append(None)
            ind_coefs.append(None)
            
        elif len(axis) > 1 and ii in axis[1:]:
            ind_coefs.append(None)
        else:
            shape_pts.append(coefs.shape[ii])
            ind_coefs.append(jj)
            ind_x.append(jj)
            jj += 1
    
    # ----------------
    # shape_other
    
    shape_other = tuple([
        ss for ii, ss in enumerate(coefs.shape)
        if ii not in axis
    ])
    
    return tuple(shape_pts), shape_other, axis_x, ind_coefs, ind_x


def _get_slices_iter(
    axis=None,
    axis_x=None,
    shape_coefs=None,
    shape_x=None,
    ind=None,
    ind_coefs=None,
    ind_x=None,
):
    
    sli_c = tuple([
        slice(None) if ii in axis else ind[ind_coefs[ii]]
        for ii in range(len(shape_coefs))
    ])

    sli_x =tuple([
        slice(None) if ii in axis_x else ind[ind_x[ii]]
        for ii in range(len(shape_x))
    ])
    
    return sli_c, sli_x


def _get_slices_iter_reverse_coefs(
    axis=None,
    shape_coefs=None,
    ind_coefs=None,
    return_as_list=None,
):
    
    sli_c = [
        ind_coefs if ii in axis else slice(None)
        for ii in range(len(shape_coefs))
    ]
    
    if return_as_list is True:
        return sli_c
    else:
        return tuple(sli_c)


def _get_slices_iter_reverse_x(
    axis_x=None,
    shape_x=None,
    ind_x=None,
):

    sli_x =tuple([
        ind_x if ii in axis_x else slice(None)
        for ii in range(len(shape_x))
    ])
    
    return sli_x


def _get_slice_out(indout=None, axis=None, shape_coefs=None):
    
    if len(axis) == 1:
        sli_out = tuple([
            indout if ii == axis[0] else slice(None)
            for ii in range(len(shape_coefs))
        ])
    else:
        sli_out = tuple([
            indout if ii == axis[0] else slice(None)
            for ii in range(len(shape_coefs))
            if ii not in axis[1:]
        ])
    
    return sli_out


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 13 14:41:08 2023

@author: dvezinet
"""

# Built-in
import os


# Common
import numpy as np
import datastock as ds
import bsplines2d as bs2


__all_ = ['load_eqdsk']


# ########################################################
# ########################################################
#               EQDSK
# ########################################################


def load_eqdsk(
    dpfe=None,
    returnas=None,
    strict=None,
    # keys
    kmesh=None,
    # optional time
    t=None,
    ktime=None,
    knt=None,
    t_units=None,
):
    """ load multiple eqdsk equilibria files and concatenate them

    If provided, vector t is used as the time vector
    Otherwise t is just a range

    Parameters
    ----------
    dpfe : str of dict of {path: [patterns]}
        DESCRIPTION. The default is None.
    returnas : dict, True or Collection
        DESCRIPTION. The default is None.
    kmesh : str
        DESCRIPTION. The default is None.
    t : sequence, optional
        DESCRIPTION. The default is None.

    Raises
    ------
    error
        DESCRIPTION.
    Exception
        DESCRIPTION.

    Returns
    -------
    out : dict or Collection
        DESCRIPTION.

    """

    # --------------------
    # check inputs
    # --------------------

    lpfe, returnas, coll, kmesh, t, ktime, knt, t_units, geqdsk = check_inputs(
        dpfe=dpfe,
        returnas=returnas,
        kmesh=kmesh,
        strict=strict,
        # optional time
        t=t,
        ktime=ktime,
        knt=knt,
        t_units=t_units,
    )

    # ----------------
    # load and extract
    # ----------------

    # loop on all files
    dfail = {}
    npfe = len(lpfe)
    for ii, pfe in enumerate(lpfe):

        # --------------
        # open and load

        with open(pfe, "r") as ff:
            data = geqdsk.read(ff)

        # ----------
        # initialize

        if ii == 0:
            # extract nb of knots
            nR = data['nx']
            nZ = data['ny']

            # extract R
            R = data['rleft'] + np.linspace(0, data['rdim'], nR)

            # extract Z
            Z = data['zmid'] + 0.5 * data['zdim'] * np.linspace(-1, 1, nZ)

            # initialize psi
            psi = np.full((npfe, nR, nZ), np.nan)

        # ------------
        # safety check

        c0 = (
            data['nx'] == nR
            and data['ny'] == nZ
            and data['rleft'] == R[0]
            and data['zmid'] == 0.5*(Z[0] + Z[-1])
        )
        if not c0:
            dfail[pfe] = f"({data['nx']}, {data['ny']})"

        # ---------------
        # extract psi map

        psi[ii, :, :] = data['psi']

    # -------------
    # sort vs time

    if t is None:
        psi = psi[0, :, :]

    # ----------------------
    # raise error if needed

    if len(dfail) > 0:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dfail.items()]
        msg = (
            "The following files have unmatching (R, Z) grids:\n"
            f"\t- reference from {lpfe[0]}: ({nR}, {nZ})\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # -------------------
    # build output
    # -------------------

    if returnas is dict:
        ddata, dref = _to_dict(
            R=R,
            Z=Z,
            psi=psi,
            t=t,
            knt=knt,
            ktime=ktime,
            t_units=t_units,
        )

    else:
        _to_Collection(
            coll=coll,
            kmesh=kmesh,
            R=R,
            Z=Z,
            psi=psi,
            # optional time
            t=t,
            ktime=ktime,
            knt=knt,
        )

    # -------------------
    # return
    # -------------------

    if returnas is dict:
        out = ddata, dref
    elif returnas is True:
        out = coll
    else:
        out = None

    return out


# ########################################################
# ########################################################
#               check
# ########################################################


def check_inputs(
    dpfe=None,
    returnas=None,
    strict=None,
    # keys
    kmesh=None,
    t=None,
    knt=None,
    ktime=None,
    t_units=None,
):

    # --------------------
    # check dependency
    # --------------------

    # import dependency
    try:
        from freeqdsk import geqdsk
    except Exception as err:
        msg = (
            "loading an eqdsk file requires an optional dependency:\n"
            "\t- file trying to load: {pfe}\n"
            "\t- required dependency: freeqdsk"
        )
        err.args = (msg,)
        raise err

    # --------------------
    # check pfe
    # --------------------

    lpfe = ds.get_files(
        dpfe=dpfe,
        returnas=list,
        strict=strict,
    )

    # --------------------
    # check returnas
    # --------------------

    if returnas is None:
        returnas = True

    coll = None
    if returnas is True:
        from ._class10_Inversion import Inversion as Collection
        coll = Collection()

    elif issubclass(returnas.__class__, bs2.BSplines2D):
        coll = returnas

    elif returnas is not dict:
        msg = (
            "returnas must be either:\n"
            "\t- dict: return mesh in ddata and dref dict\n"
            "\t- True: return mesh in new Collection instance\n"
            "\t- Collection instance: add mesh 2d rect to it, named kmesh"
        )
        raise Exception(msg)

    # -------------------
    # kmesh
    # -------------------

    if coll is not None:
        wm = coll._which_mesh
        kmesh = ds._generic_check._obj_key(
            d0=coll.dobj.get(wm, {}),
            short='m',
            key=kmesh,
            ndigits=2,
        )

    # -------------------
    # time
    # -------------------

    if len(lpfe) > 1:

        # time vector
        if t is None:
            t = np.arange(0, len(lpfe))
        t = ds._generic_check._check_flat1darray(
            t, 't',
        )

        # key ref time
        knt = ds._generic_check._obj_key(
            d0=coll.dref,
            short='nt',
            key=knt,
            ndigits=2,
        )

        # key time
        ktime = ds._generic_check._obj_key(
            d0=coll.ddata,
            short='t',
            key=ktime,
            ndigits=2,
        )

        # t_units
        t_units = ds._generic_check._check_var(
            t_units, 't_units',
            types=str,
            default='s',
        )

    else:
        t, ktime, knt, t_units = None, None, None, None

    return lpfe, returnas, coll, kmesh, t, ktime, knt, t_units, geqdsk


# ########################################################
# ########################################################
#               to dict and Collection
# ########################################################


def _to_dict(
    R=None,
    Z=None,
    psi=None,
    t=None,
    knt=None,
    ktime=None,
    t_units=None,
):

    nR = R.size
    nZ = Z.size

    # ref keys
    knR = 'nR'
    knZ = 'nZ'
    kpsi = 'psi'

    # ref
    dref = {
        'nR': {'size': nR},
        'nZ': {'size': nZ},
    }
    if t is not None:
        dref[knt] = {'size': t.size}
        ref = (knt, knR, knZ)
    else:
        ref = (knR, knZ)

    # data keys
    kR = 'R'
    kZ = 'Z'
    kpsi = 'psi2d'

    ddata = {
        kR: {
            'data': R,
            'ref': (knR,),
            'units': 'm',
        },
        kZ: {
            'data': Z,
            'ref': (knZ,),
            'units': 'm',
        },
        kpsi: {
            'data': psi,
            'ref': ref,
            'units': '',
        },
    }
    if t is not None:
        ddata[ktime] = {
            'data': t,
            'ref': knt,
            'units': t_units,
        }

    return ddata, dref


def _to_Collection(
    coll=None,
    kmesh=None,
    R=None,
    Z=None,
    psi=None,
    # optional time
    t=None,
    ktime=None,
    knt=None,
    t_units=None,
):

    # add mesh
    coll.add_mesh_2d_rect(
        key=kmesh,
        knots0=R,
        knots1=Z,
        deg=1,
    )

    # add time
    if t is not None:
        coll.add_ref(key=knt, size=t.size)
        coll.add_data(
            key=ktime,
            data=t,
            ref=knt,
            units=t_units,
        )

    # add psi2d
    kbs = f"{kmesh}_bs1"
    ref = kbs if t is None else (knt, kbs)

    coll.add_data(
        key='psi2d',
        data=psi,
        ref=ref,
        units='',
    )

    # # add rhopn2d
    psi0 = np.nanmin(psi, axis=(-2, -1))
    if t is not None:
        rhopn2d = (psi0[:, None, None] - psi) / psi0[:, None, None]
    else:
        rhopn2d = (psi0 - psi) / psi0

    coll.add_data(
        key='rhopn2d',
        data=rhopn2d,
        ref=ref,
        units='',
    )


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: cdivision=True
#
from libc.stdlib cimport malloc, free

cdef ChainedList* create_ordered(int N) nogil:
   # Create a vector of size N, s.t. vecotr = {0, 1, 2, ..., N-1}
   cdef int ii
   cdef ChainedList* head
   cdef ChainedList* current
   cdef ChainedList* new_node

   # empty chained list
   if N==0:
       return <ChainedList*>NULL

   # there is at least one:
   head = <ChainedList*>malloc(sizeof(ChainedList))
   head.val = 0
   head.size = 1
   head.next = NULL
   head.last = head

   if N==1:
       return head

   # there are at least 2
   current = head

   for ii in range(1, N):
       new_node = <ChainedList*>malloc(sizeof(ChainedList))
       new_node.val = ii
       new_node.next = NULL
       current.next = new_node
       current = new_node
       head.size += 1

   head.last = current
   return head


cdef void pop_at_pos(ChainedList** head, int n) nogil:
  # Pop the element at the n-th position
  cdef ChainedList* to_pop
  cdef ChainedList* previous

  if n==0:
      to_pop = head[0]
      head[0] = head[0].next
      head[0].size = to_pop.size - 1
      head[0].last = to_pop.last
      free(to_pop)
      return

  to_pop = head[0].next
  previous = head[0]
  for _ in range(1, n):
      previous = to_pop
      to_pop = to_pop.next

  previous.next = to_pop.next
  if to_pop.next == NULL:
      head[0].last = previous

  head[0].size-=1
  free(to_pop)
  return


cdef double get_at_pos(ChainedList* head, int n) nogil:
  # Get the n-th element
  cdef ChainedList* current

  current = head
  for _ in range(0, n):
      current = current.next

  return current.val


cdef void push_back(ChainedList** head, double val) nogil:
  # adding element at the end of the chain
  cdef ChainedList* new_node

  new_node = <ChainedList*>malloc(sizeof(ChainedList))
  new_node.val = val
  new_node.next = NULL

  if head[0] == NULL:
      new_node.size = 1
      new_node.last = new_node
      head[0] = new_node
      return

  head[0].last.next = new_node
  head[0].last = new_node
  head[0].size += 1
  return


cdef void free_cl(ChainedList** head) nogil:
  cdef ChainedList* current

  if head[0] == NULL :
      return

  previous = head[0]
  current = head[0].next
  while current != NULL:
      free(previous)
      previous = current
      current = current.next
  free(previous)
  return



# -*- coding: utf-8 -*-


# Built-in
# import copy


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from mpl_toolkits.mplot3d import Axes3D
import datastock as ds


# tofu
from ..geom import _comp_solidangles
from . import _class8_plot
from . import _generic_check


# ###############################################################
# ###############################################################
#                   Main
# ###############################################################


def main(
    coll=None,
    key_diag=None,
    key_cam=None,
    indch=None,
    indref=None,
    # parameters
    res=None,
    margin_par=None,
    margin_perp=None,
    config=None,
    # solid angle
    n0=None,
    n1=None,
    # lamb
    res_lamb=None,
    # bool
    verb=None,
    plot=None,
    # plotting
    indplot=None,
    dax=None,
    plot_config=None,
    fs=None,
    dmargin=None,
    vmin_cam0=None,
    vmax_cam0=None,
    vmin_cam=None,
    vmax_cam=None,
    vmin_cam_lamb=None,
    vmax_cam_lamb=None,
    vmin_plane=None,
    vmax_plane=None,
):

    # ------------
    # check inputs

    (
        key_diag, key_cam, indch, indref,
        parallel, is2d, spectro, doptics,
        lop_pre, lop_post,
        res, margin_par, margin_perp,
        verb, plot,
        indref, indplot,
        plot_config,
    ) = _check(
        coll=coll,
        key_diag=key_diag,
        key_cam=key_cam,
        indch=indch,
        indref=indref,
        # parameters
        res=res,
        margin_par=margin_par,
        margin_perp=margin_perp,
        # bool
        verb=verb,
        plot=plot,
        # plotting
        indplot=indplot,
        plot_config=plot_config,
        config=config,
    )

    # ----------
    # los_ref

    klos = doptics['los']

    # start_ref
    ptsx, ptsy, ptsz = coll.get_rays_pts(
        key=klos,
    )
    ipts = tuple(np.r_[-2, indref])
    pt_ref = np.r_[ptsx[ipts], ptsy[ipts], ptsz[ipts]]

    # los_ref
    vx, vy, vz = coll.get_rays_vect(
        key=klos,
        norm=True,
    )
    iv = tuple(np.r_[-1, indref])
    los_ref = np.r_[vx[iv], vy[iv], vz[iv]]

    # -----------------
    # furthest aperture

    if spectro:
        if len(lop_post) > 0:
            poly = lop_post[-1]
        else:
            kmax = 0.
    else:
        if len(doptics['optics']) > 0:
            poly = doptics['optics'][-1]
        else:
            kmax = 0.

    # poly => sca
    if poly is not None:
        px, py, pz = coll.get_optics_poly(poly)
        poly = np.array([px, py, pz])
        kmax = np.max(np.sum(
            (poly - pt_ref[:, None])*los_ref[:, None],
            axis=0,
        ))

    # get length along los
    klos = kmax + margin_par

    # get pt plane
    pt_plane = pt_ref + klos * los_ref

    # -------------------------------------
    # create plane perpendicular to los_ref

    if parallel:
        e0_cam = coll.dobj['camera'][key_cam]['dgeom']['e0']
        e1_cam = coll.dobj['camera'][key_cam]['dgeom']['e1']
    else:
        ke0 = coll.dobj['camera'][key_cam]['dgeom']['e0']
        ke1 = coll.dobj['camera'][key_cam]['dgeom']['e1']
        e0_cam = coll.ddata[ke0]['data'][indref]
        e1_cam = coll.ddata[ke1]['data'][indref]

    e0 = np.cross(e1_cam, los_ref)
    e0 = e0 / np.linalg.norm(e0)

    e1 = np.cross(los_ref, e0)
    e1 = e1 / np.linalg.norm(e1)

    # -------------------------------------
    # create plane perpendicular to los_ref

    # get limits of plane
    if indch is None:
        kk = vx[-1, ...]

        x0, x1, iok = _get_intersect_los_plane(
            cent=pt_plane,
            nin=los_ref,
            e0=e0,
            e1=e1,
            ptx=ptsx[-2, ...],
            pty=ptsy[-2, ...],
            ptz=ptsz[-2, ...],
            vx=vx[-1, ...],
            vy=vy[-1, ...],
            vz=vz[-1, ...],
        )

    else:
        x0 = np.r_[0]
        x1 = np.r_[0]

    # dx0, dx1
    dx0 = [
        np.nanmin(x0) - margin_perp[0],
        np.nanmax(x0) + margin_perp[0],
    ]
    dx1 = [
        np.nanmin(x1) - margin_perp[1],
        np.nanmax(x1) + margin_perp[1],
    ]

    # cerate 2d grid
    nx0 = int(np.ceil((dx0[1] - dx0[0]) / res[0])) + 2
    nx1 = int(np.ceil((dx1[1] - dx1[0]) / res[1])) + 2

    x0 = np.linspace(dx0[0], dx0[1], nx0)
    x1 = np.linspace(dx1[0], dx1[1], nx1)

    ds = (x0[1] - x0[0]) * (x1[1] - x1[0])

    x0f = np.repeat(x0[:, None], nx1, axis=1)
    x1f = np.repeat(x1[None, :], nx0, axis=0)

    # derive 3d pts
    ptsx = pt_plane[0] + x0f*e0[0] + x1f*e1[0]
    ptsy = pt_plane[1] + x0f*e0[1] + x1f*e1[1]
    ptsz = pt_plane[2] + x0f*e0[2] + x1f*e1[2]

    # ----------
    # compute

    if spectro:
        dout = _spectro(
            coll=coll,
            key_diag=key_diag,
            key_cam=key_cam,
            # pts
            ptsx=ptsx,
            ptsy=ptsy,
            ptsz=ptsz,
            # solid angle
            n0=n0,
            n1=n1,
            # res
            res_lamb=res_lamb,
        )

    else:
        dout = _nonspectro(
            coll=coll,
            key_cam=key_cam,
            doptics=doptics,
            par=parallel,
            is2d=is2d,
            # points
            # pts
            ptsx=ptsx,
            ptsy=ptsy,
            ptsz=ptsz,
            config=config,
        )

    # ------------
    # format output

    dout.update({
        'key_diag': key_diag,
        'key_cam': key_cam,
        'is2d': is2d,
        'spectro': spectro,
        'indch': indch,
        'indref': indref,
        'los_ref': los_ref,
        'pt_ref': pt_ref,
        'klos': klos,
        'e0': e0,
        'e1': e1,
        'ptsx': ptsx,
        'ptsy': ptsy,
        'ptsz': ptsz,
        'x0': x0,
        'x1': x1,
        'ds': ds,
    })

    # -------
    # plot

    if plot is True:
        _plot(
            coll=coll,
            # extra
            indplot=indplot,
            dax=dax,
            plot_config=plot_config,
            fs=fs,
            dmargin=dmargin,
            vmin_cam0=vmin_cam0,
            vmax_cam0=vmax_cam0,
            vmin_cam=vmin_cam,
            vmax_cam=vmax_cam,
            vmin_cam_lamb=vmin_cam_lamb,
            vmax_cam_lamb=vmax_cam_lamb,
            vmin_plane=vmin_plane,
            vmax_plane=vmax_plane,
            # dout
            **dout,
        )

    return dout


# ###############################################################
# ###############################################################
#                   Check
# ###############################################################


def _check(
    coll=None,
    key_diag=None,
    key_cam=None,
    indch=None,
    indref=None,
    # parameters
    res=None,
    margin_par=None,
    margin_perp=None,
    # bool
    verb=None,
    plot=None,
    # plotting
    indplot=None,
    plot_config=None,
    config=None,
):

    # ----------
    # keys

    key_diag, key_cam = coll.get_diagnostic_cam(
        key=key_diag,
        key_cam=key_cam,
    )

    if len(key_cam) > 1:
        msg = f"Please select a key_cam!\n\tkey_cam: {key_cam}"
        raise Exception(msg)

    key_cam = key_cam[0]
    spectro = coll.dobj['diagnostic'][key_diag]['spectro']
    is2d = coll.dobj['diagnostic'][key_diag]['is2d']
    parallel = coll.dobj['camera'][key_cam]['dgeom']['parallel']
    doptics = coll.dobj['diagnostic'][key_diag]['doptics'][key_cam]

    if spectro and not is2d:
        msg = "Only implemented for 2d spectro"
        raise Exception(msg)

    if is2d:
        n0, n1 = coll.dobj['camera'][key_cam]['dgeom']['shape']
    else:
        nch = coll.dobj['camera'][key_cam]['dgeom']['shape'][0]

    # -----------------
    # loptics

    if spectro:
        optics, cls_optics = coll.get_optics_cls(doptics['optics'])
        ispectro = cls_optics.index('crystal')
        lop_pre = doptics['optics'][:ispectro]
        lop_post = doptics['optics'][ispectro+1:]

    else:
        lop_pre = doptics['optics']
        lop_post = []

    # -----------------
    # indch

    if indch is not None:
        if spectro or is2d:
            indch = ds._generic_check._check_flat1darray(
                indch, 'indch',
                dtype=int,
                size=2,
                sign='>0',
            )
            indch[0] = indch[0] % n0
            indch[1] = indch[1] % n1

        else:
            indch = int(ds._generic_check._check_var(
                indch, 'indch',
                types=(float, int),
                allowed=['los', 'vos'],
            )) % nch

    # -----------------
    # indref

    if indref is not None:
        if spectro or is2d:
            indref = ds._generic_check._check_flat1darray(
                indref, 'indref',
                dtype=int,
                size=2,
                sign='>=0',
            )
            indref[0] = indref[0] % n0
            indref[1] = indref[1] % n1

        else:
            indref = int(ds._generic_check._check_var(
                indref, 'indref',
                types=(float, int),
                sign='>=0',
            )) % nch

            if indch is not None:
                indref = indch

    # ----------
    # res

    if res is None:
        res = 0.001

    if isinstance(res, (int, float)):
        res = np.r_[res, res]

    res = ds._generic_check._check_flat1darray(
        res, 'res',
        dtype=float,
        size=2,
        sign='>0.',
    )

    # -----------
    # margin_par

    margin_par = ds._generic_check._check_var(
        margin_par, 'margin_par',
        types=float,
        default=0.5,
    )

    # -----------
    # margin_perp

    if margin_perp is None:
        margin_perp = 0.02

    if isinstance(margin_perp, (float, int)):
        margin_perp = [margin_perp, margin_perp]

    margin_perp = ds._generic_check._check_flat1darray(
        margin_perp, 'margin_perp',
        dtype=float,
        size=2,
        sign='>=0',
    )

    # -----------
    # verb

    verb = ds._generic_check._check_var(
        verb, 'verb',
        types=bool,
        default=True,
    )

    # -----------
    # plot

    plot = ds._generic_check._check_var(
        plot, 'plot',
        types=bool,
        default=True,
    )

    # ------------------
    # get indref for los

    if indref is None:
        if spectro or is2d or indch is None:
            etend = coll.ddata[doptics['etendue']]['data']
            indref = np.nanargmax(etend)

            if is2d:
                n0, n1 = etend.shape
                indref = np.r_[indref // n1, indref % n1].astype(int)

        else:
            indref = indch

    # -----------
    # indplot

    if indplot is None:
        indplot = indref

    # -----------
    # plot_config

    if plot is True and plot_config is None:
        plot_config = config

    return (
        key_diag, key_cam, indch, indref,
        parallel, is2d, spectro, doptics,
        lop_pre, lop_post,
        res, margin_par, margin_perp,
        verb, plot,
        indref, indplot,
        plot_config,
    )


# ###############################################################
# ###############################################################
#              get intersection los vs plane
# ###############################################################


def _get_intersect_los_plane(
    cent=None,
    nin=None,
    e0=None,
    e1=None,
    ptx=None,
    pty=None,
    ptz=None,
    vx=None,
    vy=None,
    vz=None,
):

    # ---------------
    # prepare output

    shape = vx.shape
    kk = np.full(shape, np.nan)

    # ------------
    # compute kk

    sca0 = vx * nin[0] + vy * nin[1] + vz * nin[2]
    sca1 = (
        (cent[0] - ptx) * nin[0]
        + (cent[1] - pty) * nin[1]
        + (cent[2] - ptz) * nin[2]
    )

    iok = np.abs(sca1) > 1e-6
    kk[iok] = sca1[iok] / sca0[iok]

    # --------------
    # 3d coordinates

    xx = ptx + kk * vx
    yy = pty + kk * vy
    zz = ptz + kk * vz

    # -----------------
    # local coordinates

    dx = xx - cent[0]
    dy = yy - cent[1]
    dz = zz - cent[2]

    x0 = dx * e0[0] + dy * e0[1] + dz * e0[2]
    x1 = dx * e1[0] + dy * e1[1] + dz * e1[2]

    return x0, x1, iok


# ###############################################################
# ###############################################################
#                   non-spectro
# ###############################################################


def _nonspectro(
    coll=None,
    key_cam=None,
    doptics=None,
    par=None,
    is2d=None,
    # pts
    ptsx=None,
    ptsy=None,
    ptsz=None,
    config=None,
):

    # -----------------
    # prepare apertures

    apertures = coll.get_optics_as_input_solid_angle(doptics['optics'])

    # -----------
    # prepare det

    k0, k1 = coll.dobj['camera'][key_cam]['dgeom']['outline']
    cx, cy, cz = coll.get_camera_cents_xyz(key=key_cam)
    dvect = coll.get_camera_unit_vectors(key=key_cam)

    det = {
        'cents_x': cx,
        'cents_y': cy,
        'cents_z': cz,
        'outline_x0': coll.ddata[k0]['data'],
        'outline_x1': coll.ddata[k1]['data'],
        'nin_x': np.full(cx.shape, dvect['nin_x']) if par else dvect['nin_x'],
        'nin_y': np.full(cx.shape, dvect['nin_y']) if par else dvect['nin_y'],
        'nin_z': np.full(cx.shape, dvect['nin_z']) if par else dvect['nin_z'],
        'e0_x': np.full(cx.shape, dvect['e0_x']) if par else dvect['e0_x'],
        'e0_y': np.full(cx.shape, dvect['e0_y']) if par else dvect['e0_y'],
        'e0_z': np.full(cx.shape, dvect['e0_z']) if par else dvect['e0_z'],
        'e1_x': np.full(cx.shape, dvect['e1_x']) if par else dvect['e1_x'],
        'e1_y': np.full(cx.shape, dvect['e1_y']) if par else dvect['e1_y'],
        'e1_z': np.full(cx.shape, dvect['e1_z']) if par else dvect['e1_z'],
    }

    # -------------
    # compute

    sang = _comp_solidangles.calc_solidangle_apertures(
        # observation points
        pts_x=ptsx,
        pts_y=ptsy,
        pts_z=ptsz,
        # polygons
        apertures=apertures,
        detectors=det,
        # possible obstacles
        config=config,
        # parameters
        summed=False,
        visibility=False,
        return_vector=False,
        return_flat_pts=False,
        return_flat_det=False,
    )

    return {
        'sang0': {
            'data': sang,
            'ref': None,
            'units': 'sr',
        },
    }


# ###############################################################
# ###############################################################
#                   Spectro
# ###############################################################


def _spectro(
    coll=None,
    key_diag=None,
    key_cam=None,
    is2d=None,
    # pts
    ptsx=None,
    ptsy=None,
    ptsz=None,
    # solid angle
    n0=None,
    n1=None,
    # res
    res_lamb=None,
):

    # ------------
    # prepare

    dout = coll.get_raytracing_from_pts(
        key=key_diag,
        key_cam=key_cam,
        # mesh
        key_mesh=None,
        res_RZ=None,
        res_phi=None,
        # points
        ptsx=ptsx,
        ptsy=ptsy,
        ptsz=ptsz,
        # solid angle
        n0=n0,
        n1=n1,
        # others
        res_rock_curve=None,
        lamb0=None,
        res_lamb=res_lamb,
        plot=False,
        plot_pixels=None,
        plot_config=None,
        vmin=None,
        vmax=None,
        append=False,
    )

    dout['sang0'] = {
        'data': dout['sang0'],
        'units': 'sr',
    }
    dout['sang'] = {
        'data': dout['sang'],
        'units': 'sr',
    }
    dout['sang_lamb'] = {
        'data': dout['sang_lamb'],
        'units': 'sr.m',
    }

    return dout


# ###############################################################
# ###############################################################
#                   Plot
# ###############################################################


def _plot(
    coll=None,
    # dout
    is2d=None,
    spectro=None,
    key_diag=None,
    key_cam=None,
    indch=None,
    indref=None,
    los_ref=None,
    pt_ref=None,
    klos=None,
    e0=None,
    e1=None,
    ptsx=None,
    ptsy=None,
    ptsz=None,
    x0=None,
    x1=None,
    ds=None,
    sang0=None,
    sang=None,
    sang_lamb=None,
    # extra
    indplot=None,
    dax=None,
    plot_config=None,
    fs=None,
    dmargin=None,
    vmin_cam0=None,
    vmax_cam0=None,
    vmin_cam=None,
    vmax_cam=None,
    vmin_cam_lamb=None,
    vmax_cam_lamb=None,
    vmin_plane=None,
    vmax_plane=None,
    **kwdargs,
):

    # ------------
    # prepare

    (
        etend0, etend, etend_lamb,
        etend_plane0, etend_plane, etend_plane_lamb,
        sli,
        extent_cam, extent_plane,
        vmin_cam0, vmax_cam0,
        vmin_cam, vmax_cam,
        vmin_cam_lamb, vmax_cam_lamb,
        vmin_plane0, vmax_plane0,
        vmin_plane, vmax_plane,
        los_refr, los_refz,
    ) = _check_plot(
        coll=coll,
        key_diag=key_diag,
        key_cam=key_cam,
        is2d=is2d,
        spectro=spectro,
        sang0=sang0,
        sang=sang,
        sang_lamb=sang_lamb,
        x0=x0,
        x1=x1,
        ds=ds,
        pt_ref=pt_ref,
        los_ref=los_ref,
        indref=indref,
        indplot=indplot,
        vmin_cam=vmin_cam,
        vmax_cam=vmax_cam,
        vmin_cam0=vmin_cam0,
        vmax_cam0=vmax_cam0,
        vmin_cam_lamb=vmin_cam_lamb,
        vmax_cam_lamb=vmax_cam_lamb,
        vmin_plane=vmin_plane,
        vmax_plane=vmax_plane,
    )

    sang_plot = sang0['data'][sli].ravel()

    # --------------
    # prepare

    dplot = coll.get_diagnostic_dplot(
        key=key_diag,
        key_cam=key_cam,
        optics=None,
        elements='o',
    )

    # --------------
    # pepare dax

    if dax is None:
        dax = _get_dax(
            is2d=is2d,
            spectro=spectro,
            fs=fs,
            dmargin=dmargin,
        )

    dax = _generic_check._check_dax(dax=dax, main='cross')

    # ----------
    # points

    kax = 'cross'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # pts
        ax.scatter(
            np.hypot(ptsx, ptsy).ravel(),
            ptsz.ravel(),
            c=sang_plot,
            s=4,
            marker='.',
            vmin=vmin_plane,
            vmax=vmax_plane,
        )

        # los_ref
        ax.plot(
            los_refr,
            los_refz,
            c='k',
            ls='-',
            lw=1.,
        )

    kax = 'hor'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # pts
        ax.scatter(
            ptsx.ravel(),
            ptsy.ravel(),
            c=sang_plot,
            s=4,
            marker='.',
            vmin=vmin_plane,
            vmax=vmax_plane,
        )

        # los_ref
        ax.plot(
            pt_ref[0] + np.r_[0, los_ref[0]],
            pt_ref[1] + np.r_[0, los_ref[1]],
            c='k',
            ls='-',
            lw=1.,
        )

    kax = '3d'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # pts
        ax.scatter(
            ptsx.ravel(),
            ptsy.ravel(),
            ptsz.ravel(),
            c=sang0['data'][sli].ravel(),
            s=4,
            marker='.',
            vmin=vmin_plane,
            vmax=vmax_plane,
        )

        # los_ref
        ax.plot(
            pt_ref[0] + np.r_[0, los_ref[0]],
            pt_ref[1] + np.r_[0, los_ref[1]],
            pt_ref[2] + np.r_[0, los_ref[2]],
            c='k',
            ls='-',
            lw=1.,
        )

    # ------------------
    # integral per pixel

    kax = 'cam_plane0'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        if is2d is True:

            # plane-specific etendue
            im = ax.imshow(
                etend_plane0.T,
                extent=extent_cam,
                origin='lower',
                interpolation='nearest',
                vmin=vmin_cam0,
                vmax=vmax_cam0,
            )

            _add_marker(ax, indref, indplot)

        else:
            # per-pixel etendue
            ax.plot(
                etend0,
                ls='-',
                lw=1.,
                c='k',
                marker='.',
            )

            # plane-specific etendue
            ax.plot(
                etend_plane0,
                ls='-',
                lw=1.,
                c='b',
                marker='.',
            )

            # ref pixel
            ax.axvline(
                indref,
                c='k',
                ls='--',
                lw=1,
            )

            # plot pixel
            ax.axvline(
                indplot,
                c='g',
                ls='--',
                lw=1,
            )

    kax = 'cam_etend0'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        if is2d is True:

            # plane-specific etendue
            im0 = ax.imshow(
                etend0.T,
                extent=extent_cam,
                origin='lower',
                interpolation='nearest',
                vmin=vmin_cam0,
                vmax=vmax_cam0,
            )

            _add_marker(ax, indref, indplot)
            plt.colorbar(im, ax=[ax, dax['cam_plane0']['handle']])

    kax = 'cam_diff0'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        if is2d is True:

            # plane-specific etendue
            diff = (etend_plane0 - etend0).T
            err = np.nanmax(np.abs(diff))
            imdiff = ax.imshow(
                diff,
                extent=extent_cam,
                origin='lower',
                interpolation='nearest',
                vmin=-err,
                vmax=err,
                cmap=plt.cm.seismic,
            )

            # ref pixel
            _add_marker(ax, indref, indplot)
            plt.colorbar(imdiff, ax=ax)

    # -------------------------------------
    # integral per pixel with rocking curve

    kax = 'cam_plane'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # plane-specific etendue
        im = ax.imshow(
            etend_plane.T,
            extent=extent_cam,
            origin='lower',
            interpolation='nearest',
            vmin=vmin_cam,
            vmax=vmax_cam,
        )

        # ref pixel
        _add_marker(ax, indref, indplot)

    kax = 'cam_etend'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # plane-specific etendue
        im0 = ax.imshow(
            etend.T,
            extent=extent_cam,
            origin='lower',
            interpolation='nearest',
            vmin=vmin_cam,
            vmax=vmax_cam,
        )

        # ref pixel
        _add_marker(ax, indref, indplot)
        plt.colorbar(im, ax=[ax, dax['cam_plane']['handle']])

    kax = 'cam_diff'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # plane-specific etendue
        diff = (etend_plane - etend).T
        err = np.nanmax(np.abs(diff))
        imdiff = ax.imshow(
            diff,
            extent=extent_cam,
            origin='lower',
            interpolation='nearest',
            vmin=-err,
            vmax=err,
            cmap=plt.cm.seismic,
        )

        # ref pixel
        _add_marker(ax, indref, indplot)
        plt.colorbar(imdiff, ax=ax)

    # -----------------------------------------------
    # integral per pixel with rocking curve and dlamb

    kax = 'cam_plane_lamb'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # plane-specific etendue
        im = ax.imshow(
            etend_plane_lamb.T,
            extent=extent_cam,
            origin='lower',
            interpolation='nearest',
            vmin=vmin_cam_lamb,
            vmax=vmax_cam_lamb,
        )

        # ref pixel
        _add_marker(ax, indref, indplot)

    kax = 'cam_etend_lamb'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # plane-specific etendue
        im0 = ax.imshow(
            etend_lamb.T,
            extent=extent_cam,
            origin='lower',
            interpolation='nearest',
            vmin=vmin_cam_lamb,
            vmax=vmax_cam_lamb,
        )

        # ref pixel
        _add_marker(ax, indref, indplot)
        plt.colorbar(im, ax=[ax, dax['cam_plane_lamb']['handle']])

    kax = 'cam_diff_lamb'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # plane-specific etendue
        diff = (etend_plane_lamb - etend_lamb).T
        err = np.nanmax(np.abs(diff))
        imdiff = ax.imshow(
            diff,
            extent=extent_cam,
            origin='lower',
            interpolation='nearest',
            vmin=-err,
            vmax=err,
            cmap=plt.cm.seismic,
        )

        # ref pixel
        _add_marker(ax, indref, indplot)
        plt.colorbar(imdiff, ax=ax)

    # -------------------------------------------
    # sang per plane point for selected pixel

    kax = 'plane0'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        im = ax.imshow(
            sang0['data'][sli].T,
            extent=extent_plane,
            origin='lower',
            interpolation='nearest',
            vmin=vmin_plane0,
            vmax=vmax_plane0,
        )

        plt.colorbar(im, ax=ax, label='solid angle (sr)')

    kax = 'plane'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        im = ax.imshow(
            sang['data'][sli].T,
            extent=extent_plane,
            origin='lower',
            interpolation='nearest',
            vmin=vmin_plane,
            vmax=vmax_plane,
        )

        plt.colorbar(im, ax=ax, label='solid angle (sr)')

    # ----------
    # diag geom

    _class8_plot._plot_diag_geom(
        dax=dax,
        key_cam=key_cam,
        dplot=dplot,
        is2d=coll.dobj['diagnostic'][key_diag]['is2d'],
    )

    # -------
    # config

    if plot_config.__class__.__name__ == 'Config':

        kax = 'cross'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax, dLeg=False)

        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax, dLeg=False)

    return dax


def _check_plot(
    coll=None,
    key_diag=None,
    key_cam=None,
    is2d=None,
    spectro=None,
    sang0=None,
    sang=None,
    sang_lamb=None,
    x0=None,
    x1=None,
    ds=None,
    pt_ref=None,
    los_ref=None,
    indref=None,
    indplot=None,
    vmin_cam=None,
    vmax_cam=None,
    vmin_cam0=None,
    vmax_cam0=None,
    vmin_cam_lamb=None,
    vmax_cam_lamb=None,
    vmin_plane=None,
    vmax_plane=None,
    vmin_plane0=None,
    vmax_plane0=None,
):

    # ---------
    # data

    if is2d is True:
        # sang['data']
        pass
    else:
        pass

    # ---------
    # integral

    if spectro:
        etend_plane0 = np.nansum(np.nansum(sang0['data'], axis=-1), axis=-1) * ds
        etend_plane = np.nansum(np.nansum(sang['data'], axis=-1), axis=-1) * ds

        ketend = coll.dobj['diagnostic'][key_diag]['doptics'][key_cam]['etendue0']
        etend0 = coll.ddata[ketend]['data']
        ketend = coll.dobj['diagnostic'][key_diag]['doptics'][key_cam]['etendue']
        etend = coll.ddata[ketend]['data']

        dlamb = coll.get_diagnostic_data(
            key_diag,
            key_cam=key_cam,
            data='dlamb',
        )[0][key_cam]

        etend_lamb = etend * dlamb
        etend_plane_lamb = sang_lamb['data'] * ds

    else:
        etend_plane0 = np.nansum(np.nansum(sang0['data'], axis=-1), axis=-1) * ds
        etend_plane = None

        ketend = coll.dobj['diagnostic'][key_diag]['doptics'][key_cam]['etendue']
        etend0 = coll.ddata[ketend]['data']
        etend = None

        etend_lamb = None
        etend_plane_lamb = None

    # ---------
    # extent

    # cam
    extent_cam = None
    if is2d:
        extent_cam = None
        sli = (indplot[0], indplot[1], slice(None), slice(None))
    else:
        sli = (indplot, slice(None), slice(None))

    # extent_plane
    dx0 = 0.5*(x0[1] - x0[0])
    dx1 = 0.5*(x1[1] - x1[0])
    extent_plane = (
        x0.min() - dx0,
        x0.max() + dx0,
        x1.min() - dx1,
        x1.max() + dx1,
    )

    # ----------
    # vmin, vmax

    if vmin_cam0 is None:
        vmin_cam0 = min(
            np.nanmin(etend0), np.nanmin(etend_plane0)
        )
    if vmax_cam0 is None:
        vmax_cam0 = max(
            np.nanmax(etend0), np.nanmax(etend_plane0)
        )

    if spectro:
        if vmin_cam is None:
            vmin_cam = min(
                np.nanmin(etend), np.nanmin(etend_plane)
            )
        if vmax_cam is None:
            vmax_cam = max(
                np.nanmax(etend), np.nanmax(etend_plane)
            )
        if vmin_cam_lamb is None:
            vmin_cam_lamb = min(
                np.nanmin(etend_lamb), np.nanmin(etend_plane_lamb)
            )
        if vmax_cam_lamb is None:
            vmax_cam_lamb = max(
                np.nanmax(etend_lamb), np.nanmax(etend_plane_lamb)
            )

    else:
        vmin_cam, vmax_cam = None, None
        vmin_cam_lamb, vmax_cam_lamb = None, None

    if vmin_plane0 is None:
        vmin_plane0 = np.nanmin(sang0['data'])
    if vmax_plane0 is None:
        vmax_plane0 = np.nanmax(sang0['data'])

    if sang is not None:
        if vmin_plane is None:
            vmin_plane = np.nanmin(sang['data'])
        if vmax_plane is None:
            vmax_plane = np.nanmax(sang['data'])

    # --------
    # los_ref

    add = np.linspace(0, 1, 20)
    los_refr = np.hypot(
        pt_ref[0] + add * los_ref[0],
        pt_ref[1] + add * los_ref[1],
    )
    los_refz = pt_ref[2] + add * los_ref[2]

    return (
        etend0, etend, etend_lamb,
        etend_plane0, etend_plane, etend_plane_lamb,
        sli,
        extent_cam, extent_plane,
        vmin_cam0, vmax_cam0,
        vmin_cam, vmax_cam,
        vmin_cam_lamb, vmax_cam_lamb,
        vmin_plane0, vmax_plane0,
        vmin_plane, vmax_plane,
        los_refr, los_refz,
    )


def _get_dax(
    is2d=None,
    spectro=None,
    fs=None,
    dmargin=None,
):

    # ----------
    # check

    if fs is None:
        fs = (15, 9)

    dmargin = ds._generic_check._check_var(
        dmargin, 'dmargin',
        types=dict,
        default={
            'bottom': 0.05, 'top': 0.95,
            'left': 0.05, 'right': 0.95,
            'wspace': 0.4, 'hspace': 0.5,
        },
    )

    # --------
    # prepare

    fig = plt.figure(figsize=fs)

    if spectro is True:
        gs = gridspec.GridSpec(ncols=8, nrows=8, **dmargin)

        # --------
        # create

        # geometry
        ax0 = fig.add_subplot(gs[:2, :2], aspect='equal', adjustable='datalim')
        ax0.set_xlabel('X (m)')
        ax0.set_ylabel('Y (m)')

        ax1 = fig.add_subplot(gs[2:4, :2], aspect='equal', adjustable='datalim')
        ax1.set_xlabel('R (m)')
        ax1.set_ylabel('Z (m)')

        ax2 = fig.add_subplot(gs[4:, :2], projection='3d')
        ax2.set_xlabel('X (m)')
        ax2.set_ylabel('Y (m)')
        ax2.set_ylabel('Z (m)')

        # images for etend0
        ax30 = fig.add_subplot(gs[:2, 2:4], aspect='equal')
        ax30.set_ylabel('x1 (m)')
        ax30.set_xlabel('x0 (m)')
        ax30.set_title('etendue\nw/o rock. curve', size=12, fontweight='bold')

        ax31 = fig.add_subplot(gs[:2, 4:6], sharex=ax30, sharey=ax30)
        ax31.set_xlabel('x0 (m)')
        ax31.set_title('integral\nw/o rock. curve', size=12, fontweight='bold')

        ax32 = fig.add_subplot(gs[:2, 6:], sharex=ax30, sharey=ax30)
        ax32.set_xlabel('x0 (m)')
        ax32.set_title('difference', size=12, fontweight='bold')

        # images for etend
        ax40 = fig.add_subplot(gs[2:4, 2:4], aspect='equal')
        ax40.set_ylabel('x1 (m)')
        ax40.set_xlabel('x0 (m)')
        ax40.set_title('etendue\nwith rock. curve', size=12, fontweight='bold')

        ax41 = fig.add_subplot(gs[2:4, 4:6], sharex=ax30, sharey=ax30)
        ax41.set_xlabel('x0 (m)')
        ax41.set_title('integral\nwith rock. curve', size=12, fontweight='bold')

        ax42 = fig.add_subplot(gs[2:4, 6:], sharex=ax30, sharey=ax30)
        ax42.set_xlabel('x0 (m)')
        ax42.set_title('difference', size=12, fontweight='bold')

        # images for etend * dlamb
        ax50 = fig.add_subplot(gs[4:6, 2:4], aspect='equal')
        ax50.set_ylabel('x1 (m)')
        ax50.set_xlabel('x0 (m)')
        ax50.set_title('etendue * dlamb\nwith rock. curve', size=12, fontweight='bold')

        ax51 = fig.add_subplot(gs[4:6, 4:6], sharex=ax30, sharey=ax30)
        ax51.set_xlabel('x0 (m)')
        ax51.set_title('integral * dlamb\nwith rock. curve', size=12, fontweight='bold')

        ax52 = fig.add_subplot(gs[4:6, 6:], sharex=ax30, sharey=ax30)
        ax52.set_xlabel('x0 (m)')
        ax52.set_title('difference', size=12, fontweight='bold')

        # plane
        ax60 = fig.add_subplot(
            gs[-2:, 2:5],
            aspect='equal',
            adjustable='datalim',
        )
        ax60.set_ylabel('x1 (m)')
        ax60.set_xlabel('x0 (m)')

        ax61 = fig.add_subplot(
            gs[-2:, 5:],
            aspect='equal',
            adjustable='datalim',
        )
        ax61.set_ylabel('x1 (m)')
        ax61.set_xlabel('x0 (m)')

        # dict
        dax = {
            'cross': {'handle': ax0, 'type': 'cross'},
            'hor': {'handle': ax1, 'type': 'hor'},
            '3d': {'handle': ax2, 'type': '3d'},
            'cam_etend0': {'handle': ax30, 'type': 'camera'},
            'cam_plane0': {'handle': ax31, 'type': 'camera'},
            'cam_diff0': {'handle': ax32, 'type': 'camera'},
            'cam_etend': {'handle': ax40, 'type': 'camera'},
            'cam_plane': {'handle': ax41, 'type': 'camera'},
            'cam_diff': {'handle': ax42, 'type': 'camera'},
            'cam_etend_lamb': {'handle': ax50, 'type': 'camera'},
            'cam_plane_lamb': {'handle': ax51, 'type': 'camera'},
            'cam_diff_lamb': {'handle': ax52, 'type': 'camera'},
            'plane0': {'handle': ax60, 'type': 'misc'},
            'plane': {'handle': ax61, 'type': 'misc'},
        }

    else:
        gs = gridspec.GridSpec(ncols=5, nrows=6, **dmargin)

        # --------
        # create

        ax0 = fig.add_subplot(gs[:3, :2], aspect='equal', adjustable='datalim')
        ax0.set_xlabel('X (m)')
        ax0.set_ylabel('Y (m)')

        ax1 = fig.add_subplot(gs[3:, :2], aspect='equal', adjustable='datalim')
        ax1.set_xlabel('R (m)')
        ax1.set_ylabel('Z (m)')

        ax2 = fig.add_subplot(gs[:2, 2:], projection='3d')
        ax2.set_xlabel('X (m)')
        ax2.set_ylabel('Y (m)')
        ax2.set_ylabel('Z (m)')

        if is2d is True:
            ax30 = fig.add_subplot(gs[2:4, 2], aspect='equal')
            ax30.set_ylabel('x1 (m)')
            ax30.set_xlabel('x0 (m)')
            ax30.set_title('integral', size=12, fontweight='bold')

            ax31 = fig.add_subplot(gs[2:4, 3], sharex=ax30, sharey=ax30)
            ax31.set_xlabel('x0 (m)')
            ax31.set_title('etendue', size=12, fontweight='bold')

            ax32 = fig.add_subplot(gs[2:4, 4], sharex=ax30, sharey=ax30)
            ax32.set_xlabel('x0 (m)')
            ax32.set_title('difference', size=12, fontweight='bold')

        else:
            ax30 = fig.add_subplot(gs[2:4, 2:])
            ax30.set_ylabel('x1 (m)')

        ax4 = fig.add_subplot(gs[4:, 2:], aspect='equal', adjustable='datalim')
        ax4.set_ylabel('x1 (m)')
        ax4.set_xlabel('x0 (m)')

        # dict
        dax = {
            'cross': {'handle': ax0, 'type': 'cross'},
            'hor': {'handle': ax1, 'type': 'hor'},
            '3d': {'handle': ax2, 'type': '3d'},
            'cam_plane0': {'handle': ax30, 'type': 'camera'},
            'plane0': {'handle': ax4, 'type': 'misc'},
        }
        if is2d is True:
            dax['cam_etend0'] = {'handle': ax31, 'type': 'camera'}
            dax['cam_diff0'] = {'handle': ax32, 'type': 'camera'}

    return dax  # , cax


def _add_marker(ax=None, indref=None, indplot=None):

    # ref pixel
    ax.plot(
        [indref[0]],
        [indref[1]],
        marker='s',
        markerfacecolor='None',
        markeredgecolor='k',
        ms=4,
    )

    # plot pixel
    ax.plot(
        [indplot[0]],
        [indplot[1]],
        marker='s',
        markerfacecolor='None',
        markeredgecolor='g',
        ms=4,
    )



# -*- coding: utf-8 -*-


# Built-in
# import copy


# Common
import numpy as np


# tofu
from ._class07_Camera import Camera as Previous
from . import _class8_check as _check
from . import _class8_compute as _compute
from . import _class08_get_data as _get_data
from . import _class08_concatenate_data as _concatenate
from . import _class8_move as _move
from . import _class8_los_data as _los_data
from . import _class08_interpolate_along_los as _interpolate_along_los
from . import _class8_equivalent_apertures as _equivalent_apertures
from . import _class8_etendue_los as _etendue_los
from . import _class8_vos as _vos
from . import _class8_vos_spectro_nobin_at_lamb as _vos_nobin_at_lamb
from . import _class8_los_angles as _los_angles
from . import _class08_generate_rays as _generate_rays
from . import _class8_plane_perp_to_los as _planeperp
from . import _class8_compute_signal as _compute_signal
from . import _class8_compute_signal_moments as _signal_moments
from . import _class8_reverse_ray_tracing as _reverse_rt
from . import _class8_plot as _plot
from . import _class8_plot_vos as _plot_vos
from . import _class8_plot_coverage as _plot_coverage
from . import _class08_save2stp as _save2stp
from . import _class08_saveload_from_file as _saveload_from_file


__all__ = ['Diagnostic']


# ###############################################################
# ###############################################################
#                           Diagnostic
# ###############################################################


class Diagnostic(Previous):

    _show_in_summary = 'all'

    _dshow = dict(Previous._dshow)
    _dshow.update({
        'diagnostic': [
            'is2d',
            'spectro',
            'PHA',
            'camera',
            'signal',
        ],
    })

    def add_diagnostic(
        self,
        key=None,
        doptics=None,
        # etendue
        etendue=None,
        # config for los
        config=None,
        length=None,
        # reflections
        reflections_nb=None,
        reflections_type=None,
        key_nseg=None,
        # compute
        compute=True,
        add_points=None,
        convex=None,
        # spectro-only
        rocking_curve_fw=None,
        # others
        compute_vos_from_los=None,
        verb=None,
        **kwdargs,
    ):

        # -----------
        # adding diag

        # check / format input
        dref, ddata, dobj = _check._diagnostics(
            coll=self,
            key=key,
            doptics=doptics,
            **kwdargs,
        )
        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

        # ---------------------
        # adding etendue / los

        key = list(dobj['diagnostic'].keys())[0]
        dopt = dobj['diagnostic'][key]['doptics']
        computable = any([len(v0['optics']) > 0 for v0 in dopt.values()])
        if compute is True and computable:
            self.compute_diagnostic_etendue_los(
                key=key,
                analytical=True,
                numerical=False,
                res=None,
                check=False,
                # los
                config=config,
                length=length,
                reflections_nb=reflections_nb,
                reflections_type=reflections_type,
                key_nseg=key_nseg,
                # equivalent aperture
                add_points=add_points,
                convex=convex,
                # spectro-only
                rocking_curve_fw=rocking_curve_fw,
                # bool
                compute_vos_from_los=compute_vos_from_los,
                verb=verb,
                plot=False,
                store='analytical',
            )

    # -----------------
    # remove
    # -----------------

    def remove_diagnostic(self, key=None, key_cam=None):
        return _check._remove(
            coll=self,
            key=key,
            key_cam=key_cam,
        )

    # -----------------
    # utilities
    # -----------------

    def get_diagnostic_ref(self, key=None, key_cam=None):
        return _check.get_ref(coll=self, key=key, key_cam=key_cam)

    def get_diagnostic_cam(self, key=None, key_cam=None, default=None):
        return _check._get_default_cam(
            coll=self,
            key=key,
            key_cam=key_cam,
            default=default,
        )

    def get_diagnostic_data(
        self,
        key=None,
        key_cam=None,
        data=None,
        rocking_curve=None,
        units=None,
        default=None,
        print_full_doc=None,
        **kwdargs,
    ):
        """ Return dict of built-in data for chosen cameras

        data can be:
            'etendue'
            'amin'
            'amax'
            'tangency radius'
            'lamb'
            'lambmin'
            'lambmax'
            'res'

        """
        return _get_data.main(
            coll=self,
            key=key,
            key_cam=key_cam,
            data=data,
            rocking_curve=rocking_curve,
            units=units,
            default=default,
            print_full_doc=print_full_doc,
            **kwdargs,
        )

    def get_diagnostic_data_concatenated(
        self,
        key=None,
        key_data=None,
        key_cam=None,
        flat=None,
    ):
        """ Return concatenated data for chosen cameras


        """
        return _concatenate.main(
            coll=self,
            key=key,
            key_data=key_data,
            key_cam=key_cam,
            flat=flat,
        )

    # -----------------
    # etendue computing
    # -----------------

    def compute_diagnostic_etendue_los(
        self,
        key=None,
        # parameters
        analytical=None,
        numerical=None,
        res=None,
        check=None,
        margin_par=None,
        margin_perp=None,
        # spectro-only
        ind_ap_lim_spectral=None,
        rocking_curve_fw=None,
        rocking_curve_max=None,
        # equivalent aperture
        add_points=None,
        convex=None,
        # for storing los
        config=None,
        length=None,
        reflections_nb=None,
        reflections_type=None,
        key_nseg=None,
        # bool
        compute_vos_from_los=None,
        return_dcompute=None,
        verb=None,
        plot=None,
        store=None,
        overwrite=None,
        debug=None,
    ):
        """ Compute the etendue of the diagnostic (per pixel)

        Etendue (m2.sr) can be computed analytically or numerically
        If plot, plot the comparison between all computations
        If store = 'analytical' or 'numerical', overwrites the diag etendue

        """

        # prepare computation
        dcompute, store, overwrite = _etendue_los.compute_etendue_los(
            coll=self,
            key=key,
            # etendue
            analytical=analytical,
            numerical=numerical,
            res=res,
            check=check,
            margin_par=margin_par,
            margin_perp=margin_perp,
            # spectro-only
            ind_ap_lim_spectral=ind_ap_lim_spectral,
            rocking_curve_fw=rocking_curve_fw,
            rocking_curve_max=rocking_curve_max,
            # equivalent aperture
            add_points=add_points,
            convex=convex,
            # bool
            verb=verb,
            plot=plot,
            store=store,
            overwrite=overwrite,
            debug=debug,
        )

        # compute los angles
        c0 = (
            any([np.any(np.isfinite(v0['los_x'])) for v0 in dcompute.values()])
            and store
        )
        if c0:
            _los_angles.compute_los_angles(
                coll=self,
                key=key,
                # los
                config=config,
                length=length,
                reflections_nb=reflections_nb,
                reflections_type=reflections_type,
                key_nseg=key_nseg,
                dcompute=dcompute,
                compute_vos_from_los=compute_vos_from_los,
                overwrite=overwrite,
            )

        if return_dcompute is True:
            return dcompute

    def compute_diagnostic_solidangle_from_plane(
        self,
        key_diag=None,
        key_cam=None,
        indch=None,
        indref=None,
        # parameters
        res=None,
        margin_par=None,
        margin_perp=None,
        config=None,
        # solid angle
        n0=None,
        n1=None,
        # lamb
        res_lamb=None,
        # bool
        verb=None,
        plot=None,
        # plotting
        indplot=None,
        dax=None,
        plot_config=None,
        fs=None,
        dmargin=None,
        vmin_cam0=None,
        vmax_cam0=None,
        vmin_cam=None,
        vmax_cam=None,
        vmin_cam_lamb=None,
        vmax_cam_lamb=None,
        vmin_plane=None,
        vmax_plane=None,
    ):
        """ Creates a plane perpendicular to los
        compute contribution of each point to the signal
        """

        return _planeperp.main(
            coll=self,
            key_diag=key_diag,
            key_cam=key_cam,
            indch=indch,
            indref=indref,
            # parameters
            res=res,
            margin_par=margin_par,
            margin_perp=margin_perp,
            config=config,
            # solid angle
            n0=n0,
            n1=n1,
            # lamb
            res_lamb=res_lamb,
            # bool
            verb=verb,
            plot=plot,
            # plotting
            indplot=indplot,
            dax=dax,
            plot_config=plot_config,
            fs=fs,
            dmargin=dmargin,
            vmin_cam0=vmin_cam0,
            vmax_cam0=vmax_cam0,
            vmin_cam=vmin_cam,
            vmax_cam=vmax_cam,
            vmin_cam_lamb=vmin_cam_lamb,
            vmax_cam_lamb=vmax_cam_lamb,
            vmin_plane=vmin_plane,
            vmax_plane=vmax_plane,
        )

    # -----------------
    # add rays from diag
    # -----------------

    def add_rays_from_diagnostic(
        self,
        key=None,
        strategy=None,
        nrays=None,
        # storing
        store=None,
        config=None,
        overwrite=None,
    ):
        return _generate_rays.main(
            coll=self,
            key=key,
            strategy=strategy,
            nrays=nrays,
            # storing
            store=store,
            config=config,
            overwrite=overwrite,
        )

    # -----------------
    # solid angle from plane
    # -----------------

    def plot_diagnostic_solidangle_from_plane(
        self,
        dout=None,
        # plotting
        indplot=None,
        dax=None,
        plot_config=None,
        fs=None,
        dmargin=None,
        vmin_cam0=None,
        vmax_cam0=None,
        vmin_cam=None,
        vmax_cam=None,
        vmin_cam_lamb=None,
        vmax_cam_lamb=None,
        vmin_plane=None,
        vmax_plane=None,
    ):
        """ Creates a plane perpendicular to los
        compute contribution of each point to the signal
        """

        return _planeperp._plot(
            coll=self,
            # extra
            indplot=indplot,
            dax=dax,
            plot_config=plot_config,
            fs=fs,
            dmargin=dmargin,
            vmin_cam0=vmin_cam0,
            vmax_cam0=vmax_cam0,
            vmin_cam=vmin_cam,
            vmax_cam=vmax_cam,
            vmin_cam_lamb=vmin_cam_lamb,
            vmax_cam_lamb=vmax_cam_lamb,
            vmin_plane=vmin_plane,
            vmax_plane=vmax_plane,
            # dout
            **dout,
        )

    def compute_diagnostic_vos(
        self,
        key_diag=None,
        key_cam=None,
        key_mesh=None,
        config=None,
        # sampling
        res_RZ=None,
        res_phi=None,
        lamb=None,
        res_lamb=None,
        res_rock_curve=None,
        n0=None,
        n1=None,
        convexHull=None,
        # user-defined limits
        user_limits=None,
        # keep3d
        keep3d=None,
        return_vector=None,
        # margins
        margin_poly=None,
        # raytracing
        visibility=None,
        # spectro-only
        rocking_curve_fw=None,
        rocking_curve_max=None,
        # bool
        check=None,
        verb=None,
        debug=None,
        store=None,
        overwrite=None,
        replace_poly=None,
        timing=None,
    ):
        """ Compute the vos of the diagnostic (per pixel)

        - poly_margin (0.3) fraction by which the los-estimated vos is widened
        -store:
            - if replace_poly, will replace the vos polygon approximation
            - will store the toroidally-integrated solid angles

        Return dvos, dref

        """

        return _vos.compute_vos(
            coll=self,
            key_diag=key_diag,
            key_cam=key_cam,
            key_mesh=key_mesh,
            config=config,
            # etendue
            res_RZ=res_RZ,
            res_phi=res_phi,
            lamb=lamb,
            res_lamb=res_lamb,
            res_rock_curve=res_rock_curve,
            n0=n0,
            n1=n1,
            convexHull=convexHull,
            # user-defined limits
            user_limits=user_limits,
            # keep3d
            keep3d=keep3d,
            return_vector=return_vector,
            # margins
            margin_poly=margin_poly,
            # spectro-only
            rocking_curve_fw=rocking_curve_fw,
            rocking_curve_max=rocking_curve_max,
            # bool
            visibility=visibility,
            check=check,
            verb=verb,
            debug=debug,
            store=store,
            overwrite=overwrite,
            replace_poly=replace_poly,
            timing=timing,
        )

    def check_diagnostic_dvos(
        self,
        key=None,
        key_cam=None,
        dvos=None,
    ):
        """ Check dvos and return it if stored """
        return _vos._check_get_dvos(
            coll=self,
            key=key,
            key_cam=key_cam,
            dvos=dvos,
        )

    def get_dvos_xyz(
        self,
        key=None,
        key_cam=None,
        dvos=None,
    ):
        """ Return ptsx, ptsy, ptsz from 3d vos """
        return _vos.get_dvos_xyz(
            coll=self,
            key_diag=key,
            key_cam=key_cam,
            dvos=dvos,
        )

    def store_diagnostic_vos(
        self,
        key_diag=None,
        dvos=None,
        dref=None,
        spectro=None,
        overwrite=None,
        replace_poly=None,
    ):
        """ Store a pre-computed dvos """
        _vos._store(
            coll=self,
            key_diag=key_diag,
            dvos=dvos,
            dref=dref,
            spectro=spectro,
            overwrite=overwrite,
            replace_poly=replace_poly,
        )

    def compute_diagnostic_vos_nobin_at_lamb(
        self,
        key_diag=None,
        key_cam=None,
        key_mesh=None,
        lamb=None,
        config=None,
        # parameters
        res_RZ=None,
        res_phi=None,
        res_rock_curve=None,
        n0=None,
        n1=None,
        convexHull=None,
        # margins
        margin_poly=None,
        nmax_rays=None,
        # spectro-only
        rocking_curve_fw=None,
        rocking_curve_max=None,
        # optional binning
        dobin=None,
        bin0=None,
        bin1=None,
        remove_raw=None,
        # bool
        visibility=None,
        verb=None,
        debug=None,
        # plot
        plot=None,
        pix0=None,
        pix1=None,
        tit=None,
    ):
        """ Compute the image of a mono-wavelength plasma volume

        Does ray-tracing from the whole plasma volume
        For a set of discrete user-defined wavelengths
        Optionally (dobin=True) also provides the binned (per-pixel) image

        - lamb: float or sequence of floats, wavelength to be traced

        Parameters for plasma volume sampling:
            - res_RZ: float (m)
            - res_phi: float (m)
            - res_rock_curve: float (rad)

        Parameters for sampling the solid angle for each point source
            - n0: int, nb of rays in horizontal direction
            - n1: int, nb of rays in vertical direction

        Parameters for binning (by default uses the camera pixels)
            - dobin: bool, do the binning or not
            - bin0: int, nb of pixels in the horizontal direction
            - bin1: int, nb of pixels in the vertical direction

        Parameters for plotting
            - plot: bool
            - pix0: index 0 of pixels outlines to be plotted
            - pix1: index 1 of pixels outlines to be plotted

        """

        return _vos_nobin_at_lamb.compute_vos_nobin_at_lamb(
            coll=self,
            key_diag=key_diag,
            key_cam=key_cam,
            key_mesh=key_mesh,
            lamb=lamb,
            config=config,
            # etendue
            res_RZ=res_RZ,
            res_phi=res_phi,
            res_rock_curve=res_rock_curve,
            n0=n0,
            n1=n1,
            convexHull=convexHull,
            # margins
            margin_poly=margin_poly,
            nmax_rays=nmax_rays,
            # spectro-only
            rocking_curve_fw=rocking_curve_fw,
            rocking_curve_max=rocking_curve_max,
            # optional binning
            dobin=dobin,
            bin0=bin0,
            bin1=bin1,
            remove_raw=remove_raw,
            # bool
            visibility=visibility,
            verb=verb,
            debug=debug,
            # plot
            plot=plot,
            pix0=pix0,
            pix1=pix1,
            tit=tit,
        )

    # ---------------
    # utilities
    # ---------------

    def get_diagnostic_equivalent_aperture(
        self,
        key=None,
        key_cam=None,
        pixel=None,
        # inital contour
        add_points=None,
        # options
        ind_ap_lim_spectral=None,
        convex=None,
        harmonize=None,
        reshape=None,
        return_for_etendue=None,
        # plot
        plot=None,
        verb=None,
        store=None,
        debug=None,
    ):
        """ Get the equivalent projected aperture for each pixel
        """

        return _equivalent_apertures.equivalent_apertures(
            coll=self,
            key=key,
            key_cam=key_cam,
            pixel=pixel,
            # inital contour
            add_points=add_points,
            # options
            ind_ap_lim_spectral=ind_ap_lim_spectral,
            convex=convex,
            harmonize=harmonize,
            reshape=reshape,
            return_for_etendue=return_for_etendue,
            # plot
            plot=plot,
            verb=verb,
            store=store,
            debug=debug,
        )

    # ---------------
    # wavelneght from angle
    # ---------------

    def get_diagnostic_lamb(
        self,
        key=None,
        key_cam=None,
        lamb=None,
        rocking_curve=None,
        units=None,
    ):
        """ Return the wavelength associated to
        - 'lamb'
        - 'lambmin'
        - 'lambmax'
        - 'res' = lamb / (lambmax - lambmin)

        """
        return _compute.get_lamb_from_angle(
            coll=self,
            key=key,
            key_cam=key_cam,
            lamb=lamb,
            rocking_curve=rocking_curve,
            units=units,
        )

    # ---------------
    # utilities
    # ---------------

    def get_optics_cls(self, optics=None):
        """ Return list of optics and list of their classes

        """
        return _check._get_optics_cls(coll=self, optics=optics)

    # def get_diagnostic_doptics(self, key=None):
    #     """
    #     Get dict of optics and corresponding classes

    #     """
    #     return _check._get_diagnostic_doptics(coll=self, key=key)

    def get_optics_outline(
        self,
        key=None,
        add_points=None,
        min_threshold=None,
        mode=None,
        closed=None,
        ravel=None,
        total=None,
    ):
        """ Return the optics outline """
        return _compute.get_optics_outline(
            coll=self,
            key=key,
            add_points=add_points,
            min_threshold=min_threshold,
            mode=mode,
            closed=closed,
            ravel=ravel,
            total=total,
        )

    def get_optics_poly(
        self,
        key=None,
        add_points=None,
        min_threshold=None,
        mode=None,
        closed=None,
        ravel=None,
        total=None,
        return_outline=None,
    ):
        """ Return the optics outline """
        return _compute.get_optics_poly(
            coll=self,
            key=key,
            add_points=add_points,
            min_threshold=min_threshold,
            mode=mode,
            closed=closed,
            ravel=ravel,
            total=total,
            return_outline=return_outline,
        )

    def get_optics_as_input_solid_angle(
        self,
        keys=None,
    ):
        """ Return the optics outline """
        return _compute.get_optics_as_input_solid_angle(
            coll=self,
            keys=keys,
        )

    def set_optics_color(self, key=None, color=None):
        return _check._set_optics_color(
            coll=self,
            key=key,
            color=color,
        )

    # -----------------
    # Moving
    # -----------------

    def move_diagnostic_to(
        self,
        key=None,
        key_cam=None,
        x=None,
        y=None,
        R=None,
        z=None,
        phi=None,
        theta=None,
        dphi=None,
        # computing
        compute=None,
        # los
        config=None,
        length=None,
        reflections_nb=None,
        reflections_type=None,
        key_nseg=None,
        # equivalent aperture
        add_points=None,
        convex=None,
        # etendue
        margin_par=None,
        margin_perp=None,
        verb=None,
    ):

        if compute is None:
            compute = True

        _move.move_to(
            self,
            key=key,
            key_cam=key_cam,
            x=x,
            y=y,
            R=R,
            z=z,
            phi=phi,
            theta=theta,
            dphi=dphi,
        )

        if compute:
            self.compute_diagnostic_etendue_los(
                key=key,
                # etendue
                analytical=True,
                numerical=False,
                res=None,
                check=False,
                margin_par=margin_par,
                margin_perp=margin_perp,
                # equivalent aperture
                add_points=add_points,
                convex=convex,
                # los
                config=config,
                length=length,
                reflections_nb=reflections_nb,
                reflections_type=reflections_type,
                key_nseg=key_nseg,
                # bool
                verb=verb,
                plot=False,
                store='analytical',
            )

    # -----------------
    # computing
    # -----------------

    def compute_diagnostic_solid_angle(
        self,
        key=None,
        key_cam=None,
        # pts
        ptsx=None,
        ptsy=None,
        ptsz=None,
        # options
        config=None,
        visibility=None,
        # return
        return_vect=None,
        return_alpha=None,
    ):
        return _los_data.compute_solid_angles(
            coll=self,
            key=key,
            key_cam=key_cam,
            # pts
            ptsx=ptsx,
            ptsy=ptsy,
            ptsz=ptsz,
            # options
            config=config,
            visibility=visibility,
            # return
            return_vect=return_vect,
            return_alpha=return_alpha,
        )

    def compute_diagnostic_signal(
        self,
        key=None,
        key_diag=None,
        key_cam=None,
        # integrand
        key_integrand=None,
        # spectral ref
        key_ref_spectro=None,
        # sampling
        method=None,
        res=None,
        mode=None,
        groupby=None,
        val_init=None,
        ref_com=None,
        # signal
        brightness=None,
        spectral_binning=None,
        # vos
        dvos=None,
        # verb
        verb=None,
        # timing
        timing=None,
        # store
        store=None,
        # return
        returnas=None,
    ):
        """ Compute synthetic signal for a diagnostic and an emissivity field

        """

        return _compute_signal.compute_signal(
            coll=self,
            key=key,
            key_diag=key_diag,
            key_cam=key_cam,
            # integrand
            key_integrand=key_integrand,
            # spectral ref
            key_ref_spectro=key_ref_spectro,
            # sampling
            method=method,
            res=res,
            mode=mode,
            groupby=groupby,
            val_init=val_init,
            ref_com=ref_com,
            # signal
            brightness=brightness,
            spectral_binning=spectral_binning,
            # vos
            dvos=dvos,
            # verb
            verb=verb,
            # timing
            timing=timing,
            # store
            store=store,
            # return
            returnas=returnas,
        )

    # -----------------------
    # ray-tracing from plasma
    # -----------------------

    def get_raytracing_from_pts(
        self,
        # diag
        key=None,
        key_cam=None,
        # mesh sampling
        key_mesh=None,
        res_RZ=None,
        res_phi=None,
        # pts coordinates
        ptsx=None,
        ptsy=None,
        ptsz=None,
        # res rays
        n0=None,
        n1=None,
        # optional lamb
        lamb0=None,
        res_lamb=None,
        rocking_curve=None,
        res_rock_curve=None,
        # options
        append=None,
        plot=None,
        dax=None,
        plot_pixels=None,
        plot_config=None,
        vmin=None,
        vmax=None,
        aspect3d=None,
        elements=None,
        colorbar=None,
    ):
        """ Get rays from plasma points to camera for a spectrometer diag """

        return _reverse_rt._from_pts(
            coll=self,
            # diag
            key=key,
            key_cam=key_cam,
            # mesh sampling
            key_mesh=key_mesh,
            res_RZ=res_RZ,
            res_phi=res_phi,
            # pts coordinates
            ptsx=ptsx,
            ptsy=ptsy,
            ptsz=ptsz,
            # res rays
            n0=n0,
            n1=n1,
            # optional lamb
            lamb0=lamb0,
            res_lamb=res_lamb,
            rocking_curve=rocking_curve,
            res_rock_curve=res_rock_curve,
            # options
            append=append,
            plot=plot,
            dax=dax,
            plot_pixels=plot_pixels,
            plot_config=plot_config,
            vmin=vmin,
            vmax=vmax,
            aspect3d=aspect3d,
            elements=elements,
            colorbar=colorbar,
        )

    # ---------------------
    # interpolate along los
    # ---------------------

    def interpolate_along_los(
        self,
        key_diag=None,
        key_cam=None,
        key_integrand=None,
        key_coords=None,
        # sampling
        res=None,
        mode=None,
        segment=None,
        radius_max=None,
        concatenate=None,
        # interpolating
        domain=None,
        val_out=None,
        # plotting
        vmin=None,
        vmax=None,
        plot=None,
        dcolor=None,
        dax=None,
    ):
        """ Compute and plot interpolated data along the los of the diagnostic

        """
        return _interpolate_along_los.main(
            coll=self,
            key_diag=key_diag,
            key_cam=key_cam,
            key_integrand=key_integrand,
            key_coords=key_coords,
            # sampling
            res=res,
            mode=mode,
            segment=segment,
            radius_max=radius_max,
            concatenate=concatenate,
            # interpolating
            domain=domain,
            val_out=val_out,
            # plotting
            vmin=vmin,
            vmax=vmax,
            plot=plot,
            dcolor=dcolor,
            dax=dax,
        )

    # -----------------
    # data moments
    # -----------------

    def compute_diagnostic_binned_data(
        self,
        key_diag=None,
        key_cam=None,
        # data to be binned
        data=None,
        # binning dimension
        bins0=None,
        bins1=None,
        bin_data0=None,
        bin_data1=None,
        # store
        store=None,
        # plotting
        plot=None,
    ):

        return _signal_moments.binned(
            coll=self,
            key_diag=key_diag,
            key_cam=key_cam,
            # data to be binned
            data=data,
            # binning dimension
            bins0=bins0,
            bins1=bins1,
            bin_data0=bin_data0,
            bin_data1=bin_data1,
            # store
            store=store,
            # plotting
            plot=plot,
        )

    # -----------------
    # plotting
    # -----------------

    def get_diagnostic_dplot(
        self,
        key=None,
        key_cam=None,
        optics=None,
        elements=None,
        vect_length=None,
        dx0=None,
        dx1=None,
        default=None,
    ):
        """ Return a dict with all that's necessary for plotting

        If no optics is provided, all are returned

        elements indicate, for each optics, what should be represented:
            - 'o': outline
            - 'v': unit vectors
            - 's': summit ( = center for non-curved)
            - 'c': center (of curvature)
            - 'r': rowland circle / axis of cylinder

        returned as a dict:

        dplot = {
            'optics0': {
                'o': {
                    'x0': ...,
                    'x1': ...,
                    'x': ...,
                    'y': ...,
                    'z': ...,
                    'r': ...,
                },
                'v': {
                    'x': ...,
                    'y': ...,
                    'z': ...,
                    'r': ...,
                },
            },
        }

        """

        return _compute._dplot(
            coll=self,
            key=key,
            key_cam=key_cam,
            optics=optics,
            elements=elements,
            vect_length=vect_length,
            dx0=dx0,
            dx1=dx1,
            default=default,
        )

    def plot_diagnostic(
        self,
        # keys
        key=None,
        key_cam=None,
        keyZ=None,
        # options
        optics=None,
        elements=None,
        proj=None,
        los_res=None,
        # data plot
        data=None,
        units=None,
        cmap=None,
        vmin=None,
        vmax=None,
        alpha=None,
        dx0=None,
        dx1=None,
        # config
        plot_config=None,
        plot_colorbar=None,
        # figure
        dax=None,
        dmargin=None,
        fs=None,
        wintit=None,
        # interactivity
        color_dict=None,
        nlos=None,
        dinc=None,
        connect=None,
    ):

        return _plot._plot_diagnostic(
            coll=self,
            # keys
            key=key,
            key_cam=key_cam,
            keyZ=keyZ,
            # options
            optics=optics,
            elements=elements,
            proj=proj,
            los_res=los_res,
            # data plot
            data=data,
            units=units,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            alpha=alpha,
            dx0=dx0,
            dx1=dx1,
            # config
            plot_config=plot_config,
            plot_colorbar=plot_colorbar,
            # figure
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            # interactivity
            color_dict=color_dict,
            nlos=nlos,
            dinc=dinc,
            connect=connect,
        )

    def plot_diagnostic_vos(
        self,
        key=None,
        key_cam=None,
        indch=None,
        indlamb=None,
        optics=None,
        elements=None,
        proj=None,
        los_res=None,
        # data plot
        dvos=None,
        units=None,
        cmap=None,
        vmin=None,
        vmax=None,
        vmin_tot=None,
        vmax_tot=None,
        vmin_cam=None,
        vmax_cam=None,
        dvminmax=None,
        alpha=None,
        plot_colorbar=None,
        # config
        plot_config=None,
        # figure
        dax=None,
        dmargin=None,
        fs=None,
        wintit=None,
        # interactivity
        color_dict=None,
    ):

        return _plot_vos._plot_diagnostic_vos(
            coll=self,
            key=key,
            key_cam=key_cam,
            indch=indch,
            indlamb=indlamb,
            optics=optics,
            elements=elements,
            proj=proj,
            los_res=los_res,
            # data plot
            dvos=dvos,
            units=units,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            vmin_tot=vmin_tot,
            vmax_tot=vmax_tot,
            vmin_cam=vmin_cam,
            vmax_cam=vmax_cam,
            dvminmax=dvminmax,
            alpha=alpha,
            plot_colorbar=plot_colorbar,
            # config
            plot_config=plot_config,
            # figure
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            # interactivity
            color_dict=color_dict,
        )

    # --------------------------
    # plot geometrical converage
    # --------------------------

    def plot_diagnostic_geometrical_coverage(
        self,
        key=None,
        # mesh sampling
        key_mesh=None,
        res_RZ=None,
        nan0=None,
        # plotting options
        plot_config=None,
        dcolor=None,
        dax=None,
        fs=None,
        dmargin=None,
        tit=None,
        cmap=None,
        vmin=None,
        vmax=None,
    ):
        """ Plot the geometrical coverage of a diagnostic, in a cross-section


        Parameters
        ----------
        key : str, optional
            key to the diagnostic
        key_mesh : str, optional
            key to the mesh used for sampling the cross-section
        res_RZ : float / list, optional
            resolution for sampling the cross-section. The default is None.
        plot_config : Config, optional
            DESCRIPTION. The default is None.
        dcolor : dict, optional
            dict of color, per camera

        Returns
        -------
        TYPE
            DESCRIPTION.

        """

        return _plot_coverage.main(
            coll=self,
            key=key,
            # mesh sampling
            key_mesh=key_mesh,
            res_RZ=res_RZ,
            nan0=nan0,
            # plotting options
            config=plot_config,
            dcolor=dcolor,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            dax=dax,
            fs=fs,
            dmargin=dmargin,
            tit=tit,
        )

    # --------------------------
    # save to file
    # --------------------------

    def save_diagnostic_to_file(
        # ---------------
        # input from tofu
        self,
        key=None,
        key_cam=None,
        key_optics=None,
        # ---------------
        # options
        factor=None,
        color=None,
        # ---------------
        # saving
        pfe_save=None,
        overwrite=None,
    ):
        """ Save desired diagnostic to a json or stp file

        Parameters
        ----------
        key : str, optional
            key to the existing set of rays
        pfe : str, optional
            valid path-file-extension (file str) where to save the file

        Returns
        -------
        None.

        """

        return _saveload_from_file.save(
            # ---------------
            # input from tofu
            coll=self,
            key=key,
            key_cam=key_cam,
            key_optics=key_optics,
            # ---------------
            # options
            factor=factor,
            color=color,
            # ---------------
            # saving
            pfe_save=pfe_save,
            overwrite=overwrite,
        )

    def add_diagnostic_from_file(
        self,
        pfe=None,
        returnas=False,
    ):
        """ Adds a diagnostic instance (and necessary optics) from json file

        Parameters
        ----------
        pfe :     str
            path/file.ext to desired file
        returnas: bool
            whether to return the Collection instance

        Returns
        -------
        coll:    Collection instance
            optional

        """

        return _saveload_from_file.load(
            coll=self,
            pfe=pfe,
            returnas=returnas,
        )